# -*- coding: utf-8 -*-

#The MIT License (MIT)
#
#Copyright (c) 2014 Lukasz Mentel
#
#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:
#
#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.
#
#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.

'''
Module for handling Gamess-US related jobs:
- Gamess          : running and submitting jobs, writing inputs,
- GamessInpParser : parsing the input file,
- GamessLogParser : parsing the output file,
- GamessDatParser : parsing data from the gamess PUNCH (.dat) file
'''

from __future__ import print_function

import argparse
import itertools
import math
import os
import re
import sys
from copy import copy
from subprocess import call
import numpy as np

from .calculator import Calculator, InputTemplate
from ..parsetools import slicebetween, sliceafter, parsepairs, getlines


class GamessUS(Calculator):

    '''Container object for Gamess-us jobs.'''

    def __init__(self, name="GamessUS", version="00", runopts=None, **kwargs):
        super(GamessUS, self).__init__(**kwargs)
        self.name = name
        self.inpext = '.inp'
        self.rungms = self.executable
        self.gmspath = os.path.dirname(self.executable)
        self.version = version
        self.runopts = runopts

        if os.path.isfile(os.path.join(self.gmspath, 'ddikick.x')):
            self.ddikick = os.path.join(self.gmspath, 'ddikick.x')

    @property
    def version(self):
        'Return the version.'
        return self._version

    @version.setter
    def version(self, value):
        'Check if the version exist and set it if it does.'

        versions = []
        for item in os.listdir(self.gmspath):
            match = re.match(r'[a-z]+\.([0-9]+)\.x', item)
            if match:
                versions.append(match.group(1))

        if value in versions:
            self._version = value
        else:
            raise IOError('GamessUS version {0:s} not found in {1:s}'.format(value, self.gmspath))

    @property
    def runopts(self):
        'Return the runopts'
        return self._runopts

    @runopts.setter
    def runopts(self, value):
        'Set the runopts dict'

        if value is None:
            self._runopts = ['1']
        else:
            self._runopts = value

    def get_command(self, inpfile):
        'Return the command to execute'

        return [self.executable, os.path.splitext(inpfile)[0], self.version] +\
            self.runopts

    def remove_dat(self, inpfile):
        '''
        Remove the gamess dat file if it exists in the scratch directory.
        '''

        datfile = os.path.splitext(inpfile)[0] + ".dat"
        if os.path.exists(os.path.join(self.scratch, datfile)):
            os.remove(os.path.join(self.scratch, datfile))

    def run(self, inpfile, logfile=None, remove_dat=True):
        '''
        Run a single gamess job interactively - without submitting to the
        queue.
        '''

        if remove_dat:
            self.remove_dat(inpfile)

        if logfile is None:
            logfile = os.path.splitext(inpfile)[0] + ".log"

        with open(logfile, 'w') as fobj:
            call(self.get_command(inpfile), stdout=fobj, stderr=fobj)

        return logfile

    def run_multiple(self, inputs):
        pass

    def accomplished(self, outfile):
        '''
        Return True if Gamess(US) job finished without errors.
        '''

        parser = GamessLogParser(outfile)
        return parser.accomplished()

    def parse(self, output, objective, regexp=None):
        '''
        Parser GAMESS(US) output file to get the objective.
        '''

        parser = GamessLogParser(output)

        if objective == "hf total energy":
            return parser.get_hf_total_energy()
        elif objective == "cisd total energy":
            energies = parser.get_energy_components('ci')
            return energies["TOTAL ENERGY"]
        elif objective == "correlation energy":
            energies = parser.get_energy_components('ci')
            return energies["TOTAL ENERGY"] - parser.get_hf_total_energy()
        elif objective == "regexp":
            return parser.get_variable(regexp)
        else:
            raise ValueError("unknown objective in prase {0:s}".format(objective))

    def write_input(self, fname, template=None, mol=None, basis=None,
                    core=None):
        '''
        Write the molpro input to "fname" file based on the information from
        the keyword arguments.

        Args:
            mol : :py:class:`chemtools.molecule.Molecule`
                Molecule object instance
            basis : dict or :py:class:`BasisSet <chemtools.basisset.BasisSet>`
                An instance of :py:class:`BasisSet <chemtools.basisset.BasisSet>`
                class or a dictionary of :py:class:`BasisSet <chemtools.basisset.BasisSet>`
                objects with element symbols as keys
            core : list of ints
                Molpro core specification
            template : :py:class:`str`
                Template of the input file
            fname : :py:class:`str`
                Name of the input file to be used
        '''

        out = ' $data\nbasis optimization\n{0:s}'.format(mol.symmetry)
        if mol.symmetry.lower() == 'c1':
            out += '\n'
        else:
            out += '\n\n'

        for i, atom in enumerate(mol.atoms):
            if i in mol.unique_labels:
                atombasis = basis[atom.symbol]
                out += atom.gamess_rep()
                out += atombasis.to_gamessus()

        out += ' $end'

        subs = {'basis': out, 'core': core}
        temp = InputTemplate(template)

        with open(fname, 'w') as finp:
            finp.write(temp.substitute(subs))

    def __repr__(self):
        return "\n".join([
            "<Gamess(",
            "\tname={},".format(self.name),
            "\tgmspath={},".format(self.gmspath),
            "\trungms={},".format(self.rungms),
            "\tversion={},".format(self.version),
            "\tscratch={},".format(self.scratch),
            "\trunopts={},".format(str(self.runopts)),
            ")>\n"])


class GamessInput(object):
    '''
    A class for parsing and writing gamess-us input files.
    '''

    def __init__(self, fname=None, parsed=None):
        '''
        Initialize the class.
        '''

        self.fname = fname
        self.end = " $end\n"
        self.parsed = parsed
        # not nested groups of input blocks (not parsed into a dict of dicts)
        self._notnested = ["$data", "$vec", "$ecp"]

    @property
    def parsed(self):
        'Return the `parsed` attribute'
        return self._parsed

    @parsed.setter
    def parsed(self, value):
        if isinstance(value, dict):
            self._parsed = value
        elif value is None:
            self._parsed = {}
        else:
            raise TypeError("expected a dictionary but got {0:s}".format(type(value)))

    def parse(self):
        '''
        Parse gamess input file into a dictionary of dictionaries, where the
        highest level entries are gamess namelist fileds and that contain
        dictionaries of options. All key are converted to lowercase.
        '''

        with open(self.fname, 'r') as finp:
            contents = finp.read()
        return self.parse_from_string(contents)

    def parse_from_string(self, inpstr):
        '''
        Parse gamess input file into a dictionary of dictionaries, where the
        highest level entries are gamess namelist fileds and that contain
        dictionaries of options. All key are converted to lowercase. For
        example if the following input was parsed:
        '''

        pat = re.compile(r'(?P<block>\$[a-zA-Z]{3,6})\s+(?P<entries>.*?)\$END',
                         flags=re.DOTALL | re.IGNORECASE)

        iterator = pat.finditer(inpstr)
        for match in iterator:
            if match.group("block").lower() not in self._notnested:
                self.parsed[match.group("block").lower()] = {}
                fields = [s.strip() for s in match.group("entries").split("\n")]
                for field in fields:
                    if not field.startswith("!"):
                        for line in field.split():
                            key, value = line.split("=")
                            self.parsed[match.group("block").lower()][key.lower()] = value
            elif match.group("block").lower() == "$data":
                self.parsed["$data"] = self.parse_data(match.group("entries"))
            elif match.group("block").lower() in ["$vec", "$ecp"]:
                self.parsed[match.group("block").lower()] = match.group("entries")
        return True

    def parse_data(self, datastr, parse_basis=False):
        '''
        Parse $DATA block specified in the gamess input file. The parser
        assumes that in $DATA input the atom positions are specified using
        cartesian coordiantes and returns a list of dictionaries with parsed
        information about each specified atom.

        Args:
            datastr (str)
                string with the contents of the $data block

        Returns:
            datadict (dict)
        '''

        block = re.compile(r'(?P<label>[a-zA-Z]{1,2}[0-9]{0,2})\s*' +
                           r'(?P<atomic>\d+\.\d+)' +
                           r'(?P<xyz>(\s+\-?\d+\.\d+){3})\s*' +
                           r'(?P<basis>.*?)\n\s*\n', flags=re.S)

        datadict = dict()

        datadict["title"] = datastr.split('\n')[0].strip()
        datadict["group"] = datastr.split('\n')[1].strip()
        datadict["atoms"] = list()

        itfound = block.finditer(datastr)
        for match in itfound:
            datadict["atoms"].append({
                'label': match.group('label'),
                'atomic': float(match.group('atomic')),
                'xyz': tuple(float(x) for x in match.group('xyz').split()),
                'basis': match.group('basis'),
            })

        return datadict

    def parsed2str(self):
        '''
        Return a string with formatted string containing the GAMESS(US) input
        based on the previously parsed data.
        '''

        inpstr = ""
        # write nested namelist groups
        for key, value in sorted(self.parsed.items()):
            if key not in self._notnested:
                inpstr += " {0:<s}\n".format(key)
                for kkey, vvalue in sorted(value.items()):
                    inpstr += "    {k:s}={v:s}\n".format(k=kkey, v=str(vvalue))
                inpstr += self.end
        # write $data card
        inpstr += self.data2str()
        return inpstr

    def data2str(self, header=True):
        '''
        Return the $DATA card of the input as a formatted string based on the
        previously parsed data.a
        '''

        data = " {0:s}\n".format("$data")
        if header:
            data += "{0:s}\n".format(self.parsed["$data"]["title"])
            data += "{0:s}\n".format(self.parsed["$data"]["group"])
            if self.parsed['$data']['group'].lower() != 'c1':
                data += '\n'
        for atom in self.parsed['$data']['atoms']:
            data += '{0:5s}{1:5.1f}{2:12.5f}{3:12.5f}{4:12.5f}\n'.format(
                atom['label'], atom['atomic'], atom['xyz'][0], atom['xyz'][1],
                atom['xyz'][2])
            data += atom['basis'] + '\n\n'
        data += self.end
        return data

    def write_input(self, filename):
        '''
        Write a gamess input file under the name <inpfile> based on the
        information fstored in the dictionary <inpdict>.

        Args:
          filename : str
            Name of the input file
        '''

        with open(filename, "w") as finp:
            finp.write(self.parsed2str())

    def write_with_vec(self, filename, vecstr, keepscftyp=False):
        '''
        Write new gamess input based on exisiting input and natural orbitals
        from $VEC section in PUNCH file.


        Write a new input file for gamess based on previously prased and/or
        constructed dictionary containing input specification and append
        orbitals from a previous run as starting orbitals. The starting
        orbitals should be stored in an ASCII PUNCH file whose name is given
        under "datfile" variable.

        Args:
            filename : str
                Name of the input file to be created
            vecstr : str
                String with the vectors to be written
        '''

        parsed = copy(self.parsed)

        if not keepscftyp:
            self.parsed['$contrl']['scftyp'] = 'none'
        self.parsed['$guess'] = {}
        self.parsed['$guess']['guess'] = 'moread'
        self.parsed['$guess']['norb'] = str(get_naos_nmos(vecstr)[1])

        with open(filename, "w") as finp:

            finp.write(self.parsed2str())

            finp.write("\n $vec\n")
            finp.write(vecstr)
            finp.write("\n $end\n")
            finp.close()

        self.parsed = parsed

    def set_gamess_input(self, dinp, mol, bs, code, core):

        if "$contrl" in dinp.keys():
            dinp["$contrl"]["icharg"] = mol.charge
            dinp["$contrl"]["mult"] = mol.multiplicity
            if mol.multiplicity == 1:
                dinp["$contrl"]["scftyp"] = "rhf"
            elif mol.multiplicity > 1:
                dinp["$contrl"]["scftyp"] = "rohf"
            if code["method"].lower() == "hf":
                dinp["$contrl"] = {key: value for key, value in dinp["$contrl"].items() if key != "cityp"}
        else:
            sys.exit("no $contrl group in the gamess input string")
        if "$cidet" in dinp.keys():
            dinp["$cidet"]["nact"] = bs.get_no_functions(bs) - core
            dinp["$cidet"]["ncore"] = core
            dinp["$cidet"]["nels"] = mol.electrons - core*2
        if "$ormas" not in dinp.keys():
            dinp["$ormas"] = {}
        if code["method"] == "cisd":
            # sum of singly and doubly occupied  orbitals
            noccupied = mol.multiplicity - 1 + (mol.electrons-(mol.multiplicity - 1))/2
            dinp["$ormas"]["nspace"] = 2
            dinp["$ormas"]["mstart(1)"] = ','.join([str(core+1),str(noccupied+1)])
            dinp["$ormas"]["mine(1)"] = ",".join([str((mol.electrons-core*2)-2), str(0)])
            dinp["$ormas"]["maxe(1)"] = ",".join([str(mol.electrons-core*2), str(2)])
        elif code["method"] == "mrcisd":
            dinp["$ormas"]["nspace"] = 3
            dinp["$ormas"]["mstart(1)"] = ','.join([str(core+1),str(mol.electrons-core+1)])
            dinp["$ormas"]["mine(1)"] = ",".join([str(mol.electrons-2), str(0)])
            dinp["$ormas"]["maxe(1)"] = ",".join([str(mol.electrons), str(2)])

        return dinp

    def print_parsed(self):
        '''
        Neat print of the dictionary with parsed input
        '''

        for key, value in sorted(self.parsed.items()):
            if isinstance(value, dict):
                print(key)
                for kkey, vvalue in sorted(value.items()):
                    print("\t{0:<10s} : {1:}".format(kkey, vvalue))
            else:
                print(key, '\n', value)


class GamessLogParser(object):
    '''
    Methods for parsing gamess-us log file.
    '''

    def __init__(self, log):
        self.logfile = log

    @property
    def logfile(self):
        'Return the value of the `logfile` attribute'
        return self._logfile

    @logfile.setter
    def logfile(self, value):
        'Check if the logfile exists first.'
        if os.path.exists(value):
            self._logfile = value
        else:
            raise ValueError("File: {} does not exist".format(value))

    def logexists(self):
        '''
        Check if the log file exists.
        '''

        if os.path.exists(self.logfile):
            return True
        else:
            sys.exit("Gamess log file: {0:s} doesn't exist in {1:s}".format(
                self.logfile, os.getcwd()))

    def parse(self, regex):
        '''
        Parsing function based on querries in the form of regular expressions
        that return the match object.
        '''

        cpatt = re.compile(regex)
        with open(self.logfile, 'r') as log:
            content = log.read()
        return cpatt.search(content)

    def accomplished(self):
        '''Check if a job teminated normally.'''

        regex = r'TERMINATED NORMALLY'
        match = self.parse(regex)
        return match is not None

    def get_version(self):
        '''
        Get the version of the GAMESS(US) package.
        '''

        regex = r'.*\s+GAMESS VERSION =\s*(.*?)\s*\*'
        match = self.parse(regex)
        if match:
            return match.group(1)
        else:
            return None

    def get_charge(self):
        '''Get total charge.'''

        regex = r'CHARGE OF MOLECULE\s+=\s*(?P<charge>\d+)'
        match = self.parse(regex)
        if match:
            return int(match.group("charge"))

    def get_electrons(self):
        '''Get number of electrons.'''

        regex = r'NUMBER OF ELECTRONS\s+=\s*(?P<nele>\d+)'
        match = self.parse(regex)
        if match:
            return int(match.group("nele"))

    def get_homo(self):
        '''
        Get the orbital index of homo orbital (indexing starts from zero).
        '''

        if int(self.get_electrons()) % 2 == 0:
            return int(self.get_electrons()) / 2 - 1
        else:
            sys.exit("open shell handling not implemented")

    def get_number_of_atoms(self):
        '''Get total number of atoms from gamess log file.'''

        regex = r'TOTAL NUMBER OF ATOMS\s+=\s*(?P<nat>\d+)'
        match = self.parse(regex)
        if match:
            return int(match.group("nat"))

    def get_number_of_aos(self):
        '''
        Get the number of primitive cartesian gaussian basis functions from
        Gamess log file
        '''

        regex = r'NUMBER OF CARTESIAN GAUSSIAN BASIS FUNCTIONS =\s*(?P<nao>\d+)'
        match = self.parse(regex)
        if match:
            return int(match.group("nao"))

    def get_number_of_core_mos(self):
        '''
        Get the number of core molecular orbitals from the Gamess log file
        '''

        regex = r'NUMBER OF CORE MOLECULAR ORBITALS\s*=\s*(?P<ncore>\d+)'
        match = self.parse(regex)
        if match:
            return int(match.group("ncore"))

    def get_number_of_mos(self):
        '''
        Get the number of molecular orbitals from Gammess log file.
        '''

        ispher_patt = r'ISPHER=\s*(?P<ispher>\-?\d{1}).*'
        var_space_patt = r'.*VARIATION SPACE IS\s*(?P<nmo>\d+).*'
        c_ispher = re.compile(ispher_patt)
        c_var_space = re.compile(var_space_patt)

        with open(self.logfile, 'r') as log:
            lines = log.read()

        match = c_ispher.search(lines)
        if match:
            ispher = int(match.group("ispher"))

        if ispher == -1:
            return self.get_number_of_aos()
        elif ispher == 1:
            match = c_var_space.search(lines)
            if match:
                n_mo = int(match.group("nmo"))
                return n_mo
        else:
            sys.exit("wrong ispher found: {0:d}".format(ispher))

    def get_linear_deps(self):
        '''
        Get number of linearly dependent combinations dropped.
        '''

        regex = r'NUMBER OF LINEARLY DEPENDENT MOS DROPPED=\s*(?P<lindep>\d+)'
        match = self.parse(regex)
        if match:
            return int(match.group('lindep'))

    def get_scf_type(self):
        '''
        Get the information on SCFTYP used in the gamess job.
        '''

        regex = r'.*SCFTYP=(?P<scftyp>[A-Z]+)'
        match = self.parse(regex)
        if match:
            return match.group("scftyp")

    def get_cc_type(self):
        '''
        Get the information on CCTYP used in the gamess job.
        '''

        regex = r'.*CCTYP =(?P<cctyp>[A-Z\(\)\-]+)'
        match = self.parse(regex)
        if match:
            return match.group("cctyp")
        else:
            return None

    def get_ci_type(self):
        '''
        Get the information on CITYP used in the gamess job.
        '''

        regex = r'.*CITYP =(?P<cityp>[A-Z]+)'
        match = self.parse(regex)
        if match:
            return match.group("cityp")

    def get_mplevel(self):
        '''
        Get the information on MPLEVL used in the gamess job.
        '''

        regex = r'.*MPLEVL=\s*(?P<mplevl>\d+)'
        match = self.parse(regex)
        if match:
            return int(match.group("mplevl"))

    def get_hf_total_energy(self):
        '''
        Return the total HF energy.
        '''

        regex = r'FINAL R[O]*HF ENERGY IS\s+(?P<energy>\-?\d+\.\d+)'
        match = self.parse(regex)
        if match:
            return float(match.group("energy"))

    def get_cc_total_energy(self):
        '''
        Independently of CCTYP value return the "HIGHEST LEVEL RESULT" total
        energy.
        '''

        regex = r'COUPLED-CLUSTER ENERGY E(.*?)\s*=\s*(?P<energy>\-?\d+\.\d+)'
        match = self.parse(regex)
        if match:
            return float(match.group("energy"))

    def get_energy_components(self, method):
        '''
        Read the summary of the energies printed in the gamess log file at the
        property section corresponding to a particular "method".
        '''

        if method.lower() in ["hf", "scf", "hfscf"]:
            if self.get_scf_type() != "NONE":
                header = 'PROPERTY VALUES FOR THE {0:<5s} SELF-CONSISTENT FIELD WAVEFUNCTION'.format(self.get_scf_type())
            else:
                sys.exit("No HF calculation was performed, check the log file once again.")
        elif method.lower() in ["ci"]:
            if self.get_ci_type().lower() in ["guga", "ormas", "fsoci"]:
                header = '{0:<5s} CI PROPERTIES'.format(self.get_ci_type())
        else:
            sys.exit("Wrong method in <get_energy_components>: {0:s}".format(method))

        with open(self.logfile, 'r') as log:
            data = log.readlines()

        return parsepairs(sliceafter(data, header, 22))

    def get_lz_values(self):
        '''
        Get values of Lz operator for every orbital and translate them into
        labels of orbitals.
        '''

        mos = r'\s*MO\s*(?P<index>\d*)\s*\(\s*(?P<shell>\d+)\s*\)'
        weights = r'\s*HAS LZ\(WEIGHT\)=\s*(?P<lz>-?\d*\.\d+)'
        lz_patt = re.compile(mos + weights)

        locstr = self.get_loc_strings('lz values')
        lines = getlines(self.logfile, locstr)

        res = list()
        for line in lines:
            match = lz_patt.search(line)
            if match:
                index = int(match.group('index')) - 1
                shell = int(match.group('shell'))
                lz = int(abs(float(match.group('lz'))))
                res.append({"index": index, "shell": shell, "lz": lz})
        return res

    def get_ao_labels(self, orbs='hf orbs'):
        '''
        Retrieve the information about atomic basis set
        '''

        indsym = r'\s*(?P<index>\d+)\s{2}(?P<symbol>[a-zA-Z]{1,2})'
        cencom = r'\s*(?P<center>\d+)\s*(?P<component>[A-Z]{1,})'
        ao_patt = re.compile(indsym + cencom)

        istart = 3
        if orbs == 'local orbs':
            istart = 1
        locstr = self.get_loc_strings(orbs)
        lines = getlines(self.logfile, locstr)
        lines = lines[istart:istart + self.get_number_of_aos()]

        res = list()
        for line in lines:
            match = ao_patt.search(line)
            if match:
                index = int(match.group('index')) - 1
                symbol = match.group('symbol')
                center = int(match.group('center'))
                component = match.group('component')
                res.append({
                    "index": index,
                    "symbol": symbol,
                    "center": center,
                    "component": component})
        return res

    def get_variable(self, rawstring):
        'wrapper around a regex search method'

        with open(self.logfile, 'r') as out:
            data = out.read()

        genre = re.compile(rawstring, flags=re.M)
        match = genre.search(data)
        if match:
            return float(match.group(1))

    def get_loc_strings(self, name):

        locators = {
            'csfs'      : [('DETERMINANT CONTRIBUTION TO CSF', 2),
                           (' ...... END OF -DRT- GENERATION ......', -5)],
            'ci coeffs' : [('      CSF      COEF    OCCUPANCY (IGNORING CORE)', 2),
                           (' ...... END OF CI-MATRIX DIAGONALIZATION ......', 0)],
            'initial orbs' : [('INITIAL GUESS ORBITALS', 3),
                              (' ...... END OF INITIAL ORBITAL SELECTION ......', 0)],
            'hf orbs'   : [('EIGENVECTORS', 3),
                           (' ...... END OF {} CALCULATION ......'.format(self.get_scf_type()), 0)],
            'ci orbs'   : [('NATURAL ORBITALS IN ATOMIC ORBITAL', 3),
                           (' ...... END OF DENSITY MATRIX CALCULATION ......', 0)],
            'local orbs': [('LOCALIZED ORBITALS', 3),
                           (' LOCALIZATION ...', 0)],
            'lz values' : [('LZ VALUE ANALYSIS FOR', 2),
                           ('EIGENVECTORS', -2)],
        }

        if name in locators.keys():
            return locators[name]
        else:
            raise ValueError('wrong "name", should be one of: {}'.format(', '.join(locators.keys())))

    def get_orbital_labels(self, orbs, minlength=5):
        '''Return the labels of the specified orbitals

        Args:
          orbs : str
            String key for extracting locators
          minlength : int
            Minimal length of a line to be kept

        Returns:
          out : list
        '''

        locstr = self.get_loc_strings(orbs)
        lines = getlines(self.logfile, locstr)
        lines = [l for l in lines if len(l) > minlength]

        nao = self.get_number_of_aos()

        inds = list(itertools.chain(*[l.split() for l in lines[::nao + 3]]))
        evs = list(itertools.chain(*[l.split() for l in lines[1::nao + 3]]))
        syms = list(itertools.chain(*[l.split() for l in lines[2::nao + 3]]))

        return zip(inds, evs, syms)

    def get_ci_coeffs(self):
        '''
        Parse CI coefficients from a list of lines containing the output of
        GAMESS(US) calculation

        Returns:
          coeffs : dict
            Dictionary with parsed coeficients and their indexes in the
            internal GAMESS(US) ordering.
        '''

        locstr = self.get_loc_strings('ci coeffs')
        lines = getlines(self.logfile, locstr)

        patt = re.compile(r'\s*(\d+)\s*(-?\d*\.\d+)')
        index = []
        coeff = []
        for line in lines:
            match = patt.search(line)
            if match:
                index.append(match.group(1))
                coeff.append(match.group(2))

        return dict(zip([int(x) for x in index], [float(x) for x in coeff]))

    def get_csfs(self, withcoeffs=True):
        '''
        Parse CSF information from a list of lines from GAMESS(US) output.
        '''

        locstr = self.get_loc_strings('csfs')
        lines = getlines(self.logfile, locstr)

        ne = self.get_electrons() - 2 * self.get_number_of_core_mos()

        csfre = r'^\s*(CSF\s*(?P<csfno>\d+)\:)?'
        coeffre = r'\s*C\(\s*(?P<detno>\d+)\)=\s*(?P<coeff>-?\d*\.\d+)'
        orbsre = r'\s+\:\s{2}(?P<orbs>.*)'
        patt = re.compile(csfre + coeffre + orbsre)

        csfstr = "".join(lines).lstrip()
        csfs = [s for s in re.split(r'CASE VECTOR =\s*\d+\s*', csfstr) if len(s) > 0]

        out = []
        for csfstr in csfs:
            endvec = csfstr.find('FOR MS=')
            vector = csfstr[:endvec].strip()
            csf = {'vector': vector}
            csf['dets'] = []
            csfstart = csfstr.find('CSF')
            for i, line in enumerate(csfstr[csfstart:].split('\n')):
                match = patt.search(line)
                if match:
                    gd = match.groupdict()
                    # in the first line of the csf specification there is the CSF number
                    # (index) that should be stored
                    if i == 0:
                        csf['no'] = int(gd['csfno'])
                        spatial = [abs(int(x)) for x in detsplit(gd['orbs'], ne)]
                        isort = np.argsort(np.array(spatial))
                        csf['spatial'] = [spatial[i] for i in isort]
                    spin = det_to_spin_coupling(gd['orbs'], ne)
                    csf['dets'].append((int(gd['detno']), float(gd['coeff']), "".join([spin[i] for i in isort])))
            out.append(csf)

        if withcoeffs:
            coeffs = self.get_ci_coeffs()
            merged = [dict(csf.items() + [('coeff', coeffs.get(csf['no'], None))]) for csf in out]
            return merged
        else:
            return out


def detsplit(det, norb):
    '''
    Split a string representation of the spatial part of the determinant `det`
    into `n` parts.

    Args:
        det : str
            String representation of the determinant
        norb : int
            Number of orbitals in the determinant

    Returns:
        out : list
            List of determinant components
    '''

    out = det.split()
    if len(out) != norb:
        nchar = int(math.floor(len(det) / float(norb)))
        out = [det[nchar * i: nchar * i + nchar] for i in range(norb)]
    return out


def det_to_spin_coupling(det, norb):
    '''
    Convert a string representaiton of a determinant to a string with spin
    coupling.

    It is assumed here that a minus sign means beta spin. `a` will denote
    :math:`\alpha` and `b` will denote :math:`\beta`

    Args:
      det : str
        String representation of a determinant, e.g. `-3  4  2 -1`
      norb : int
        Number of orbitals that should be present in `det` equal to the number
        of active electrons

    Returns:
      sc : str
        String representation of a spin coupling
    '''

    sc = ''

    detspl = detsplit(det, norb)

    for oi in detspl:
        if '-' in oi:
            sc += 'b'
        else:
            try:
                sc += 'a'
            except:
                sc += '*'
    return sc


class GamessDatParser(object):
    'Parser for the GAMESS(US) dat (.F10) file'

    def __init__(self, datfile):
        self.datfile = datfile

    @property
    def datfile(self):
        'Return the value of the `datfile` attribute'
        return self._datfile

    @datfile.setter
    def datfile(self, value):
        if os.path.exists(value):
            self._datfile = value
        else:
            raise ValueError("File: {} does not exist".format(value))

    def get_occupations(self):
        '''
        Parse the occupation numbers from the ascii PUNCH file (.dat).
        '''

        with open(self.datfile, 'r') as dat:
            data = dat.read()

        no_patt = re.compile(r'\$OCC(NO)?(?P<occ>.*?)\$END', flags=re.DOTALL)
        match = no_patt.search(data)
        nooc = list()
        if match:
            for line in match.group("occ").split('\n'):
                nooc.extend([float(x) for x in line.split()])
            return np.asarray(nooc)
        else:
            raise ValueError('No section with occupation numbers found.')

    def parse_data(self):
        '''
        Parse $DATA block from the dat file.
        '''

        with open(self.datfile, 'r') as dat:
            datastr = slicebetween(dat.read(), '$DATA', '$END')
        gip = GamessInput()
        return gip.parse_data(datastr.lstrip(' \n'))

    def get_vec_string(self, method):
        '''
        Parse the natural orbitals from the ascii PUNCH file (.dat).

        Args:
          method : str
            acceptable values are:
                - for scf orbitals: "scf", "hf", "rhf", "rohf", "uhf", "gvb"
                - for mcscf orbitals: "mcscfmos", "mcscfnos"
                - for ci orbitals: "ci", "aldet", "fsoci", "guga", "genci",
                  "ormas""
                - for localized orbitals: `local`

        Returns:
            orbs : numpy.array
              2D numpy array of shape (naos, nmos) containing ortbials
              expansion coefficients
        '''

        with open(self.datfile, 'r') as dat:
            data = dat.read()

        if method.lower() in ['scf', 'hf', 'rhf', 'rohf', 'uhf', 'gvb']:
            try:
                orbi = data.index('ORBITALS ---')
            except:
                raise ValueError('SCF orbitals header not found, check dat file')
        elif method.lower() in ['ci', 'aldet', 'fsoci', 'guga', 'genci', 'ormas']:
            try:
                orbi = data.index('NO-S')
            except:
                raise ValueError('CI orbitals header not found, check dat file')
        elif method.lower() == 'mcscfmos':
            try:
                orbi = data.index('OPTIMIZED MCSCF MO-S')
            except:
                raise ValueError('MCSCF orbitals header not found, check dat file')
        elif method.lower() == 'mcscfnos':
            try:
                orbi = data.index('NATURAL ORBITALS OF MCSCF')
            except:
                raise ValueError('MCSCF orbitals header not found, check dat file')
        elif method.lower() == 'local':
            try:
                orbi = data.index('LOCALIZED ORBITALS')
            except:
                raise ValueError('LOCALIZED orbitals header not found, check dat file')
        else:
            raise ValueError("Don't know what to do with: '{0:s}'".format(method))
        vecstr = slicebetween(data[orbi:], '$VEC', '$END').strip(" \n\t\r")
        if vecstr == "":
            raise ValueError("No $VEC section found for method: '{0:s}'".format(method))
        else:
            # add a single whitespace to account for the removed space in strip
            return ' ' + vecstr

    def get_orbitals(self, method):
        '''
        Parse the natural orbitals from the ascii PUNCH file (.dat).

        Args:
          method : str
            acceptable values are:
                - for scf orbitals: "scf", "hf", "rhf", "rohf", "uhf", "gvb"
                - for mcscf orbitals: "mcscfmos", "mcscfnos"
                - for ci orbitals: "ci", "aldet", "fsoci", "guga", "genci",
                  "ormas""
                - for localized orbitals: `local`

        Returns:
            orbs : numpy.array
              2D numpy array of shape (naos, nmos) containing ortbials
              expansion coefficients
        '''

        vecstr = self.get_vec_string(method)
        return self.parse_orbitals(vecstr)

    @staticmethod
    def parse_orbitals(vecstr, clength=15):
        '''
        Parse dat orbitals into numpy array of the shape (naos, nmos)

        Parse orbitals given a string obtained from the $VEC section of the
        PUNCH file (.dat) into a numpy array of the shape (naos, nmos) where
        "naos" and "nmos" are the number of atomic orbitals and number of
        molecular orbitals respectively. The shape is deduced from the last
        line of the string.

        Args:
          vecstr : str
            string with the contents fo the gamess $VEC block
          clength : int
            total length of the coefficient string as stored by the gamess
            format, by default gamess stores orbital coefficients in
            `e15.8` fortran format so the total length is 15.

        Returns:
          orbs : numpy.array
            2D numpy array of shape (naos, nmos) containing ortbials
            expansion coefficients
        '''

        naos, nmos, nlines = get_naos_nmos(vecstr)
        orblines = vecstr.split('\n')
        orbs = np.zeros((naos, nmos), dtype=float)
        counter = -1
        for i in range(0, nmos):
            for j in range(0, nlines):
                counter += 1
                nitems = int(len(orblines[counter][5:])) // clength
                orbs[5 * j: 5 * (j + 1), i] = \
                    [float(orblines[counter][5 + 15 * n: 5 + 15 * (n + 1)])
                     for n in range(nitems)]
        return orbs


def get_naos_nmos(vecstr, clength=15):
    '''
    Get the number of AO's and MO's from a string in $VEC block.

    Args:
        vecstr (str)
            string with the contents fo the gamess $VEC block
        clength (int)
            total length of the coefficient string as stored by the gamess
            format, by default gamess stores orbital coefficients in
            'e15.8' fortran format so the total length is 15.

    Returns:
        naos (int)
            number of atomic orbitals
        naos (int)
            number of moelcular orbitals
        nlines (int)
            number of lines per molecular orbital
    '''

    veclines = vecstr.split('\n')
    noveclines = len(veclines)
    lineit = iter(veclines)
    nlines = 0

    while next(lineit)[:2].strip() == '1':
        nlines += 1

    if nlines == 0:
        raise ValueError("'nlines' cannot be zero, check vecstr in 'get_naos_nmos'")

    naos = 5 * (nlines - 1) + len(vecstr.split('\n')[nlines - 1][5:]) // clength
    nmos = noveclines // nlines
    return naos, nmos, nlines


def to_gamess_vec(coeffs, cfmt='15.8e'):
    '''
    Given a 2D array return a string with the elements converted to the
    GAMESS(US) $VEC format

    Args:
        coeffs : numpy.array
            A two-dimensional array of type `float`

        cmft : str
            Default format for the elements of `coeffs`

    Returns:
        vec : str
            A string representation of the `coeffs` using GAMESS(US) $VEC
            format
    '''

    if coeffs.dtype != np.float64:
        raise ValueError('array should be of type np.flaot64, got: {}'.format(coeffs.dtype))

    nao, nmo = coeffs.shape

    if nao % 5 == 0:
        nlines = nao / 5
    else:
        nlines = 1 + nao / 5

    vec = ''
    for i in range(nmo):
        if i + 1 >= 100:
            ilab = i + 1 % 100
        else:
            ilab = i + 1
        for l in range(nlines):
            if l + 1 > 1000:
                llab = l + 1 % 1000
            else:
                llab = l + 1
            if l < nlines - 1:
                coeffrow = "".join(["{0:{1:}}".format(coeffs[j + l * 5, i], cfmt) for j in range(5)])
            else:
                coeffrow = "".join(["{0:{1:}}".format(coeffs[j + l * 5, i], cfmt) for j in range(0, 5 - ((l + 1) * 5 - nao))])
            vec += "{0:2d}{1:3d}{2:s}\n".format(ilab, llab, coeffrow)
    return vec


def writeorbinp():
    '''
    A script for writing the GAMESS(US) input with guess orbitals from a
    previous run.
    '''

    parser = argparse.ArgumentParser()

    parser.add_argument('inpfile')
    parser.add_argument('datfile')
    parser.add_argument('orbitals', default='ci')
    parser.add_argument('-o', '--output')
    args = parser.parse_args()

    gip = GamessInput(args.inpfile)
    gip.parse()

    gdp = GamessDatParser(args.datfile)
    vectr = gdp.get_vec_string(args.orbitals)

    if args.output is None:
        args.output = os.path.splitext(args.inpfile)[0] + '_{}.inp'.format(args.orbitals)
    gip.write_with_vec(args.output, vectr)
    print('Input written to: {}'.format(args.output))



# -*- coding: utf-8 -*-

#The MIT License (MIT)
#
#Copyright (c) 2014 Lukasz Mentel
#
#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:
#
#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.
#
#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.

import argparse
import os
import socket
import subprocess


def main():
    '''
    Script for submitting molpro job to the queue.
    '''

    parser = argparse.ArgumentParser()
    group = parser.add_mutually_exclusive_group()
    parser.add_argument("input",
                        help='molpro input file to be executed')
    parser.add_argument("-d",
                        "--dryrun",
                        action="store_true",
                        help="write the pbs script but don't submit to the queue, default-False")
    parser.add_argument("-e",
                        "--extrafiles",
                        nargs="+",
                        default=[],
                        help="additional files that need to be copied to the node/scratch")
    group.add_argument("-n",
                       "--nodes",
                       default="1",
                       help="number of nodes, default=1")
    group.add_argument("-H",
                       "--HOST",
                       default="",
                       help="destination hostname, default=''")
    parser.add_argument("-p",
                        "--ppn",
                        default="1",
                        help="number of processes per node, default=1")
    parser.add_argument("-s",
                        "--usescratch",
                        action="store_true",
                        help="use node scratch directory (/scratch/$USER)",
                        default=False)
    parser.add_argument("-q",
                        "--queue",
                        default="default",
                        help="destination queue, default=default")
    parser.add_argument("-t",
                        "--walltime",
                        default="120:00:00",
                        help="walltime in the format HH:MM:SS,\
                        default=120:00:00")
    args = vars(parser.parse_args())
    submit_pbs(args)


def set_defaults(args):
    '''Set some useful default values and add them to args'''

    args['workdir'] = os.getcwd()
    args['home'] = os.getenv("HOME")
    if socket.gethostname() == "boron.chem.umk.pl":
        args['scratch'] = os.path.join('/scratch', os.getenv('USER'))
        args['molpro'] = '/share/apps/molprop_2012_1_Linux_x86_64_i8/bin/molpro'
    else:
        args['scratch'] = os.getenv("TMPDIR")
        args['molpro'] = 'molpro'
    args['local_scr'] = os.path.join(os.getenv('HOME'), 'scratch')
    args['jobname'] = os.path.splitext(args["input"])[0]
    args['outfile'] = args['jobname'] + ".o${PBS_JOBID}"
    args['errfile'] = args['jobname'] + ".e${PBS_JOBID}"
    args['script_name'] = "run." + args['jobname']
    return args


def submit_pbs(args):
    '''
    Write the run script for PBS and submit it to the queue.
    '''

    args = set_defaults(args)

    with open(args['script_name'], 'w') as script:
        script.write("#PBS -S /bin/bash\n")
        if args['HOST'] != "":
            script.write("#PBS -l nodes={0}:ppn={1}\n".format(args['HOST'], args['ppn']))
        else:
            script.write("#PBS -l nodes={0}:ppn={1}\n".format(args['nodes'], args['ppn']))
        if "mem" in args.keys():
            script.write("#PBS -l mem={0}\n".format(args["mem"]))
        script.write("#PBS -l walltime={0}\n\n".format(args['walltime']))
        #script.write('export LD_LIBRARY_PATH=$LD_LIBRARY_PATH:/share/apps/lib\n\n')
        script.write("#PBS -o {}\n".format(args["outfile"]))
        script.write("#PBS -e {}\n".format(args["errfile"]))
        script.write("#PBS -N {}\n".format(args["jobname"]))
        if args["queue"] != "default":
            script.write("#PBS -q {}\n".format(args["queue"]))
        script.write("cd $PBS_O_WORKDIR\n")
        if args['usescratch']:
            wrkdir = os.path.join(args['scratch'], args['jobname'])
            script.write("mkdir -p {}\n".format(wrkdir))
            files = args['jobname']
            if args['extrafiles']:
                files += ' ' + ' '.join(args['extrafiles'])
            script.write('cp -t {0} {1}\n'.format(wrkdir, files))
            script.write('cd {0}\n'.format(wrkdir))
        script.write("\n{0:<s} --nobackup -o {1} {2:s} \n".format(args['molpro'],
                     args['outfile'], args['jobname']))

    # submit the job to the queue if requested
    if args['dryrun']:
        print("Created job script: {0}\n NOT submitting to the queue\nbye...".format(args['script_name']))
    else:
        print("Created job script: {0}\nsubmitting to the queue".format(args['script_name']))
        output = subprocess.check_output(["qsub", args['script_name']])
        pid = output.split(".")[0]
        return args['jobname'] + ".o" + pid

if __name__ == "__main__":
    main()




''' Chemtools setup script'''

from setuptools import setup
from setuptools.command.test import test as TestCommand
import sys


class Tox(TestCommand):

    # user_options = [('tox-args=', 'a', "Arguments to pass to tox")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.tox_args = None

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import tox
        import shlex
        args = self.tox_args
        if args:
            args = shlex.split(self.tox_args)
        errno = tox.cmdline(args=args)
        sys.exit(errno)


def readme():
    '''Return the contents of the README.rst file.'''
    with open('README.rst', encoding='utf8') as freadme:
        return freadme.read()


setup(
    author="Lukasz Mentel",
    author_email="lmmentel@gmail.com",
    description="Python tools for quantum chemical calculations",
    include_package_data=True,
    install_requires=[
        'sqlalchemy',
        'scipy >= 0.11',
        'numpy >= 1.7',
        'mendeleev',
        'numba',
        'pandas'],
    entry_points={
        'console_scripts': [
            'writeorbinp = chemtools.calculators.gamessus:writeorbinp',
            'bsconvert = chemtools.cli:bsconvert',
            'bsprint = chemtools.cli:bsprint',
        ]
    },
    license='MIT',
    long_description=readme(),
    long_description_content_type='text/x-rst',
    name='chemtools',
    packages=['chemtools', 'chemtools/pescan', 'chemtools/calculators'],
    scripts=[
        'chemtools/submitgamess.py',
        'chemtools/submitmolpro.py'],
    url='https://github.com/lmmentel/chemtools/',
    version='0.9.2',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering :: Chemistry',
        'Topic :: Scientific/Engineering :: Physics'],
    keywords=['basis set', 'optimization', 'quantum chemistry',
              'molecular physics'],
    tests_require=['tox', 'pytest'],
    cmdclass={'test': Tox},
)



# -*- coding: utf-8 -*-

#The MIT License (MIT)
#
#Copyright (c) 2014 Lukasz Mentel
#
#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:
#
#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.
#
#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.

from __future__ import division, print_function

import json
import os
import pickle
import re
from collections import OrderedDict
from copy import copy, deepcopy
from itertools import chain
import numpy as np
from scipy.linalg import sqrtm, inv
from scipy.special import factorial, factorial2, binom
from chemtools.basisparse import (parse_basis, merge_exponents, CFDTYPE, get_l,
                                  NumpyEncoder)


class BasisSet(object):
    '''
    Basis set module supporting basic operation on basis sets.

    Args:
        name : str
            Name of the basis set

        element : str
            Symbol of the element

    Kwargs:
        kind : str
            Classification of the basis set functions, *diffuse*, *tight*

        family : str
            basis set family

        functions : dict
            Dict of functions with *s*, *p*, *d*, *f*, ... as keys

        params : list of dicts
            Parameters for generating the functions according to the model
        '''

    def __init__(self, name, element, family=None, kind=None,
                 functions=None, info=None):

        self.name = name
        self.element = element
        self.family = family
        self.kind = kind
        self.functions = functions
        self.info = info

    def __add__(self, other):
        '''Add functions from another BasisSet object

        Args:
          other : BasisSet
            BasisSet object whose functions will be added to the existing ones

        Returns:
          BasisSet instance with functions from `self` and `other` merged
        '''

        newf = deepcopy(self.functions)
        for oshell, ofs in other.functions.items():
            if oshell.lower() in self.functions.keys():
                i = self.functions[oshell]['e'].size
                newf[oshell]['e'] = np.concatenate((self.functions[oshell]['e'],
                                                    ofs['e']))
                newcf = [np.copy(cf) for cf in ofs['cf']]
                for cf in newcf:
                    cf['idx'] = cf['idx'] + i
                newf[oshell]['cf'].extend(newcf)
            else:
                newf[oshell] = deepcopy(ofs)

        return BasisSet(name=self.name, element=self.element, functions=newf)

    @staticmethod
    def from_pickle(fname, **kwargs):
        '''Read a pickled BasisSet object from a pickle file

        Args:
            fname : str
                File name containing the BasisSet
            kwargs : dict
                Extra arguments for the `pickle.load` method

        Raises:
            UnicodeDecodeError
                When you try to read a python2 pickle using python3, to fix
                that use `encoding='latin1'` option
        '''

        with open(fname, 'rb') as fil:
            return pickle.load(fil, **kwargs)

    @classmethod
    def from_optpars(cls, x0, funs=None, name=None, element=None,
                     explogs=False):
        '''
        Return a basis set object generated from a sequence based on the
        specified arguments.

        Args:
          x0 : list or numpy.array
            Parameters to generate the basis set as a continuous list or array

          funs : list of tuples
            A list of tuple specifying the shell type, number of functions and
            parameters, e.g. `[('s', 'et', 4, (0.5, 2.0)), ('p', 'et', 3,
            (1.0, 3.0))]`

          name : str
            Name of the basis set

          element : str
            Chemical symbol of the element

        Returns:
          out : BasisSet
        '''

        functions = dict()
        ni = 0
        nt = 0
        for shell, seq, nf, params in funs:
            functions[shell] = dict()
            if seq in ["exp", "exponents"]:
                nt += nf
                if explogs:
                    functions[shell]['e'] = generate_exponents(seq, nf, np.exp(x0[ni: nt]))
                else:
                    functions[shell]['e'] = generate_exponents(seq, nf, x0[ni: nt])
                ni += nf
            else:
                # TODO:
                # params shouldn't be here since it can have a dummy,
                # real values are taken from x0
                nt += len(params)
                functions[shell]['e'] = generate_exponents(seq, nf, x0[ni:nt])
                ni += len(params)
        functions = OrderedDict(sorted(functions.items(),
                                key=lambda x: get_l(x[0])))
        bs = cls(name=name, element=element, functions=functions)
        bs.uncontract()
        return bs

    @classmethod
    def from_sequence(cls, funs=None, name=None, element=None):
        '''
        Return a basis set object generated from a sequence based on the
        specified arguments.

        Args:
          funs : list of tuples
            A list of tuple specifying the shell type, number of functions and
            parameters, e.g. `[('s', 'et', 4, (0.5, 2.0)), ('p', 'et', 3,
            (1.0, 3.0))]`

          name : str
            Name of the basis set

          element : str
            Chemical symbol of the element

        Returns:
          out : BasisSet
        '''

        functions = dict()

        for shell, seq, nf, params in funs:
            functions[shell] = dict()
            functions[shell]['e'] = generate_exponents(seq, nf, params)
        functions = OrderedDict(sorted(functions.items(),
                                key=lambda x: get_l(x[0])))
        bs = cls(name=name, element=element, functions=functions)
        bs.uncontract()
        bs.sort()
        return bs

    @classmethod
    def from_file(cls, fname=None, fmt=None, name=None):
        '''Read and parse a basis set from file and return a BasisSet object

        Args:
          fname : str
            File name

          fmt : str
            Format of the basis set in the file (*molpro*, *gamessus*)

          name : str
            Name of the basis set

        Returns:
          out : BasisSet or dict
            Basisset object parsed from file or dictionary of BasisSet objects
        '''

        if name is None:
            name = os.path.splitext(fname)[0]

        with open(fname, 'r') as fobj:
            basstr = fobj.read()

        return BasisSet.from_str(basstr, fmt=fmt, name=name)

    @classmethod
    def from_str(cls, string, fmt=None, name=None):
        '''
        Parse a basis set from string

        Args:
          string : str
            A string with the basis set
          fmt : str
            Format of the basis set in the file: *molpro*, *gamessus*
          name : str
            Name of the basis set

        Returns:
          out : BasisSet or dict
            Basisset object parsed from string or dictionary of BasisSet
            objects
        '''

        res = parse_basis(string, fmt=fmt)
        # return a BasisSet object if only one basis parsed or dict of BasisSet
        # objects with atomic symbol as key

        out = dict()
        if len(res) == 1:
            atom, fs = list(res.items())[0]
            return cls(name=name, element=list(res.keys())[0],
                       functions=OrderedDict(sorted(fs.items(),
                                             key=lambda x: get_l(x[0]))))
        else:
            for atom, fs in res.items():
                out[atom] = cls(name=name, element=atom,
                                functions=OrderedDict(sorted(fs.items(),
                                                      key=lambda x: get_l(x[0]))))
            return out

    @classmethod
    def from_json(cls, jsonstring, **kwargs):
        '''
        Instantiate the `BasisSet` object from a JSON string

        Args:
            jsonstring: str
                A JSON serialized string with the basis set
        '''

        def json_numpy_obj_hook(dct):
            '''
            Decodes a previously encoded numpy ndarray with proper shape
            and dtype.

            Args:
                dct: (dict) json encoded ndarray

            Returns:
                (ndarray) if input was an encoded ndarray
            '''

            if isinstance(dct, dict) and 'dtype' in dct:
                if dct['dtype'] == 'CFDTYPE':
                    return np.array([tuple(r) for r in dct['data']],
                                    dtype=CFDTYPE)
                else:
                    return np.array(dct['data'], dtype=dct['dtype'])
            return dct

        loaded = json.loads(jsonstring, object_hook=json_numpy_obj_hook,
                            **kwargs)

        return cls(**loaded)

    def append(self, other):
        '''Append functions from another BasisSet object

        Args:
          other : BasisSet
            BasisSet object whose functions will be added to the existing ones
        '''

        for oshell, ofs in other.functions.items():
            if oshell.lower() in self.functions.keys():
                i = self.functions[oshell]['e'].size
                self.functions[oshell]['e'] = np.concatenate((self.functions[oshell]['e'], ofs['e']))
                newcf = [np.copy(cf) for cf in ofs['cf']]
                for cf in newcf:
                    cf['idx'] = cf['idx'] + i
                self.functions[oshell]['cf'].extend(newcf)
            else:
                self.functions[oshell] = deepcopy(ofs)

    def completeness_profile(self, zetas):
        '''
        Calculate the completeness profile of each shell of the basis set

        Args:
          zetas : numpy.array
            Scaning exponents

        Returns:
          out : numpy.array
            numpy array with values of the profile (shells, zetas)
        '''

        out = np.zeros((zetas.size, len(self.functions.keys())))

        for i, (shell, fs) in enumerate(self.functions.items()):
            cc = self.contraction_matrix(shell)
            # calculate the overlap matrix
            S = self.shell_overlap(shell)
            # calculate the inverse square root of S
            X = inv(sqrtm(S))
            SO = primitive_overlap(get_l(shell), fs['e'], zetas)
            J = np.dot(np.dot(cc, X).T, SO)
            Y = np.sum(np.square(J), axis=0)
            out[:, i] = Y
        return out

    def contraction_matrix(self, shell):
        '''
        Return the contraction coefficients for a given shell in a
        matrix form with size `ne * nc`, where `ne` is the number of
        exponents and `nc` is the number of contracted functions

        Args:
          shell : str
            shell label, *s*, *p*, *d*, ...

        Returns:
          out : 2D numpy.array
            2D array with contraction coefficients
        '''

        if shell in self.functions.keys():
            fs = self.functions[shell]
            out = np.zeros((fs['e'].size, len(fs['cf'])))
            for i, cf in enumerate(fs['cf']):
                out[cf['idx'], i] = cf['cc']
            return out
        else:
            raise ValueError("shell '{}' is not present in the BasisSet".format(shell))

    def contractions_per_shell(self):
        '''
        Calculate how many contracted functions are in each shell.

        Returns:
          out : list of ints
        '''

        return [len(f['cf']) for s, f in self.functions.items()]

    def contraction_scheme(self):
        '''
        Return a string describing the contraction scheme.
        '''

        cs, ec = [], []
        for shell, fs in self.functions.items():
            cs.append((shell, len(fs['e']), len(fs['cf'])))
            ec.append([len(cfs) for cfs in fs['cf']])
        return "({p:s}) -> [{c:s}] : {{{ec}}}".format(
            p="".join(["{0:d}{1:s}".format(c[1], c[0]) for c in cs]),
            c="".join(["{0:d}{1:s}".format(c[2], c[0]) for c in cs]),
            ec="/".join([" ".join(["{0:d}".format(c) for c in x]) for x in ec]))

    def contraction_type(self):
        '''
        Try to determine the contraction type: segmented, general,
        uncontracted, unknown.
        '''

        pps = self.primitives_per_shell()
        ppc = self.primitives_per_contraction()

        if any(x > 1 for x in pps):
            if all(all(x == 1 for x in shell) for shell in ppc):
                return "uncontracted"
            elif all(all(pinc == np for pinc in shell)
                     for np, shell in zip(pps, ppc)):
                return "general"
            else:
                return "unknown"

        # one function per shell case
        if all(all(x == 1 for x in shell) for shell in ppc):
            return "uncontracted 1fps"

    def get_exponents(self, asdict=True):
        '''
        Return the exponents of a given shell or if the shell isn't
        specified return all of the available exponents

        Args:
            asdict (bool): if `True` a dict with exps per shell is
                           returned
        '''

        if asdict:
            return {shell: f['e'] for shell, f in self.functions.items()}
        else:
            exps = [self.functions[k]['e'] for k in self.functions.keys()]
            array = np.fromiter(chain.from_iterable(exps), dtype=np.float)
            return np.sort(array)[::-1]

    def get_filename(self, ext=None):

        name = self.name.strip().replace(' ', '_') + '-' + self.element
        if ext:
            return name + '.' + ext
        return name

    def nf(self, spherical=True):
        '''
        Calculate the number of basis functions

        Args:
          spherical : bool
            flag indicating if spherical or cartesian functions should be used,
            default: True

        Returns:
          res : int
            number of basis functions
        '''

        if spherical:
            return sum(nspherical(get_l(shell)) * len(fs['cf'])
                       for shell, fs in self.functions.items())
        else:
            return sum(ncartesian(get_l(shell)) * len(fs['cf'])
                       for shell, fs in self.functions.items())

    def normalize(self):
        '''
        Normalize contraction coefficients for each contracted functions based
        on the primitive overlaps so that the norm is equal to 1.
        '''

        for shell, fs in self.functions.items():
            cc = self.contraction_matrix(shell)
            po = primitive_overlap(get_l(shell), fs['e'], fs['e'])
            for col in range(cc.shape[1]):
                norm2 = np.dot(cc[:, col], np.dot(po, cc[:, col]))
                fs['cf'][col]['cc'] = cc[fs['cf'][col]['idx'], col] / np.sqrt(norm2)

    def normalization(self):
        '''
        For each function (contracted) calculate the norm and return a list of
        tuples containing the shell, function index and the norm respectively.
        '''

        out = list()
        for shell, fs in self.functions.items():
            cc = self.contraction_matrix(shell)
            po = primitive_overlap(get_l(shell), fs['e'], fs['e'])
            for col in range(cc.shape[1]):
                norm2 = np.dot(cc[:, col], np.dot(po, cc[:, col]))
                out.append(tuple([shell, col, norm2]))
        return out

    def nprimitive(self, spherical=True):
        '''
        Return the number of primitive functions assuming sphrical or
        cartesian gaussians.

        Args:
          spherical : bool
            A flag to select either spherical or cartesian gaussians

        Returns:
          out : int
            Number of primitive function in the basis set
        '''

        if spherical:
            # calculate the number of spherical components per shell
            ncomp = [nspherical(get_l(shell))
                     for shell in self.functions.keys()]
        else:
            # calculate the number of cartesian components per shell
            ncomp = [ncartesian(get_l(shell))
                     for shell in self.functions.keys()]

        return sum([prim * nc for prim, nc in zip(self.primitives_per_shell(),
                                                  ncomp)])

    def partial_wave_expand(self):
        '''
        From a given basis set with shells spdf... return a list of basis sets
        that are subsets of the entered basis set with increasing angular
        momentum functions included [s, sp, spd, spdf, ...]
        '''

        res = list()
        shells = self.functions.keys()
        for i in range(1, len(shells) + 1):
            bscopy = copy(self)
            bscopy.functions = {k: v for k, v in self.functions.items()
                                if k in shells[:i]}
            res.append(bscopy)
        return res

    def primitives_per_shell(self):
        '''
        Calculate how many primitive functions are in each shell.

        Returns:
          out : list of ints
        '''

        return [len(f['e']) for s, f in self.functions.items()]

    def primitives_per_contraction(self):
        '''
        Calculate how many primities are used in each contracted function.

        Returns:
          out : list of ints
        '''
        return [[len(cc) for cc in f['cf']] for s, f in self.functions.items()]

    def uncontract(self, copy=False):
        '''
        Uncontract the basis set. This replaces the contraction coefficients in
        the current object.

        Args:
          copy : bool
            If `True` return an uncontracted copy of the basis set rather than
            uncontracting in place, default is `False`.
        '''

        if copy:
            res = deepcopy(self)
            for shell, fs in res.functions.items():
                fs['cf'] = [np.array([tuple([i, 1.0])], dtype=CFDTYPE)
                            for i, _ in enumerate(fs['e'])]
            return res
        else:
            for shell, fs in self.functions.items():
                fs['cf'] = [np.array([tuple([i, 1.0])], dtype=CFDTYPE)
                            for i, _ in enumerate(fs['e'])]

    def shell_overlap(self, shell):
        '''
        Calculate the overlap integrals for a given shell

        Args:
          shell : str
            Shell

        Returns:
          out : numpy.array
            Overlap integral matrix
        '''

        exps = self.functions[shell]['e']
        S = primitive_overlap(get_l(shell), exps, exps)
        cc = self.contraction_matrix(shell)
        return np.dot(cc.T, np.dot(S, cc))

    def sort(self, reverse=False):
        '''
        Sort shells in the order of increasing angular momentum and for each
        shell sort the exponents.

        Args:
          reverse : bool
            If `False` sort the exponents in each shell in the descending order
            (default), else sort exponents in ascending order
        '''

        self.functions = OrderedDict(sorted(self.functions.items(),
                                            key=lambda x: get_l(x[0])))

        for shell, fs in self.functions.items():
            if reverse:
                idx = np.argsort(fs['e'])
            else:
                idx = np.argsort(fs['e'])[::-1]

            # actually sort the exponents and coefficients
            fs['e'] = fs['e'][idx]
            for cf in fs['cf']:
                cf['idx'] = np.asarray([np.nonzero(idx == y)[0][0]
                                        for y in cf['idx']])
                cf.sort(order='idx')

    def to_cfour(self, comment="", efmt="15.8f", cfmt="15.8f"):
        '''
        Return a string with the basis set in (new) CFOUR format.

        Args:
          comment : str
            comment string
          efmt : str
            string describing output format for the exponents,
            default: "20.10f"
          cfmt : str
            string describing output format for the contraction
            coefficients, default: "15.8f"

        Returns:
          res : str
            basis set string in Cfour format
        '''

        am, ne, cf = [], [], []
        for shell, shellfs in sorted(self.functions.items(),
                                     key=lambda x: get_l(x[0])):
            am.append(get_l(shell))
            ne.append(len(shellfs["e"]))
            cf.append(len(shellfs["cf"]))

        res = "\n{e}:{s}\n{c}\n\n".format(e=self.element, s=self.name,
                                          c=comment)
        res += "{0:3d}\n".format(len(self.functions.keys()))
        res += "".join(["{0:5d}".format(x) for x in am]) + "\n"
        res += "".join(["{0:5d}".format(x) for x in cf]) + "\n"
        res += "".join(["{0:5d}".format(x) for x in ne]) + "\n"
        res += "\n"

        for shell, fs in self.functions.items():
            for lst in splitlist(fs['e'], 5):
                res += "".join(["{0:>{efmt}}".format(e, efmt=efmt)
                                for e in lst]) + "\n"
            res += "\n"
            # create an array with all the contraction coefficients for a given shell
            cc = self.contraction_matrix(shell)
            for row in cc:
                for splitrow in splitlist(row, 5):
                    res += "{c}".format(c="".join(["{0:{cfmt}}".format(c, cfmt=cfmt) for c in splitrow])) + "\n"
            res += "\n"
        return res

    def to_dalton(self, fmt='prec'):
        '''
        Return a string with the basis set in DALTON format.

        Args:
          fmt : str
            string describing output format for the exponents and coefficents
                - `prec` "20.10f"
                - `default` "10.4f"
                - or python format string e.g. "15.8f"

        Returns:
          res : str
            basis set string in Dalton format
        '''

        formats = {'prec': '20.10f', 'default': '10.4f'}

        ffmt = formats.get(fmt, fmt)

        if fmt == 'prec':
            ffmt = '20.10f'
            nitems = 3
            fmtlabel = 'H'
        elif fmt == 'default':
            ffmt = '10.4f'
            nitems = 7
            fmtlabel = ' '
        else:
            ffmt = fmt
            cwidth = int(ffmt.split('.')[0])
            nitems = (80 - cwidth) // cwidth
            fmtlabel = "{0:d}F{1}.{2}".format(nitems, ffmt.split('.')[0], re.sub('[A-Za-z]+', '', ffmt.split('.')[1]))

        res = "! {s}\n".format(s=self.name)
        for shell, fs in self.functions.items():
            res += "! {s} functions\n".format(s=shell)
            res += "{f:1s}{p:>4d}{c:>4d}\n".format(f=fmtlabel, p=len(fs['e']),
                                                   c=len(fs['cf']))
            # create an array with all the contraction coefficients for a given shell
            cc = self.contraction_matrix(shell)
            for expt, row in zip(fs['e'], cc):
                it = splitlist(row, nitems)
                res += "{e:>{efmt}}{c}".format(e=expt, efmt=ffmt, c="".join(["{0:{cfmt}}".format(c, cfmt=ffmt) for c in next(it)])) + "\n"

                for row in it:
                    res += ' ' * int(ffmt.split('.')[0]) + \
                        "".join(["{0:{cfmt}}".format(c, cfmt=ffmt) for c in row]) + "\n"
        return res

    def to_gamessus(self, efmt="20.10f", cfmt="15.8f"):
        '''
        Return a string with the basis set in GAMESS(US) format.

        Args:
          efmt : str
            string describing output format for the exponents,
            default: "20.10f"
          cfmt : str
            string describing output format for the contraction
            coefficients, default: "15.8f"

        Returns:
          res : str
            basis set string in Gamess(US) format
        '''

        res = ''
        for shell, fs in self.functions.items():
            for contraction in fs["cf"]:
                res += "{s:<1s}{n:>3d}\n".format(s=shell.upper(),
                                                 n=len(contraction))
                for i, (idx, coeff) in enumerate(contraction, start=1):
                    res += "{i:3d}{e:>{efmt}}{c:>{cfmt}}".format(i=i,
                                                                 e=fs['e'][idx],
                                                                 efmt=efmt,
                                                                 c=coeff,
                                                                 cfmt=cfmt) + "\n"
        return res + "\n"

    def to_gaussian(self, efmt="20.10f", cfmt="15.8f"):
        '''
        Return a string with the basis set in Gaussian format.

        Args:
          efmt : str
            string describing output format for the exponents,
            default: "20.10f"
          cfmt : str
            string describing output format for the contraction
            coefficients, default: "15.8f"

        Returns:
          res : str
            basis set string in Gaussian format
        '''

        res = "****\n{e:<7s}0\n".format(e=self.element)
        for shell, fs in self.functions.items():
            for contraction in fs["cf"]:
                res += "{s:<1s}{n:>4d}{i:>7.2f}\n".format(s=shell.upper(),
                                                          n=len(contraction),
                                                          i=1.0)
                for idx, coeff in contraction:
                    res += "{e:>{efmt}}{c:>{cfmt}}".format(e=fs['e'][idx],
                                                           efmt=efmt, c=coeff,
                                                           cfmt=cfmt) + "\n"
        return res + "****\n"

    def to_json(self, fname=None, **kwargs):
        '''
        Serizalize the basis set object to JSON format
        '''

        if fname:
            with open(fname, 'w') as fjson:
                json.dump(self.__dict__, fjson, cls=NumpyEncoder, **kwargs)

        return json.dumps(self.__dict__, cls=NumpyEncoder, **kwargs)

    def to_latex(self, efmt="20.10f", cfmt="15.8f"):
        '''
        Return a string with the basis set as LaTeX table/

        Args:
            efmt : str
                Output format for the exponents, default: "20.10f"
            cfmt : str
                Output format for the contraction coefficients,
                default: "15.8f"

        Returns:
            res : str
                basis set string in LaTeX format
        '''

        # get the number of contracted functions per shell
        ncf = [sum([len(cfs) > 1 for cfs in fs['cf']])
               for sh, fs in self.functions.items()]
        idx = np.argmax(ncf)
        nccols = ncf[idx]

        out = '\\begin{{tabular}}{{{}}}\n'.format('r' * (nccols + 2))
        out += 'No. & \multicolumn{1}{c}{Exponent} & ' +\
               '\multicolumn{{{0:d}}}{{c}}{{Coefficients }} \\\ \n'.format(nccols)

        for shell, fs in self.functions.items():
            out += '\hline \n'
            out += '\multicolumn{{{0:d}}}{{c}}{{ {1:s} shell }} \\\ \hline \n'.format(nccols + 2, shell)

            cc = self.contraction_matrix(shell)
            count = 0
            # select columns with more than 1 non zero coefficients
            nonzerocolmask = np.array([np.count_nonzero(col) > 1
                                       for col in cc.T])
            nonzerorowmask = np.array([np.count_nonzero(row) > 0
                                       for row in cc[:, nonzerocolmask]])
            if np.any(nonzerocolmask):
                for expt, cfc in zip(fs['e'][nonzerorowmask],
                                     cc[np.ix_(nonzerorowmask,
                                               nonzerocolmask)]):
                    count += 1
                    coeffrow = " & ".join(["{0:{cfmt}}".format(c, cfmt=cfmt)
                                           for c in cfc])
                    out += "{i:5d} & {e:>{efmt}} & {c}".format(i=count, e=expt,
                                                               efmt=efmt,
                                                               c=coeffrow) + " \\\ \n"

            if np.any(np.logical_not(nonzerocolmask)):
                for colidx in np.where(np.logical_not(nonzerocolmask))[0]:
                    count += 1
                    nonzerorowmask = np.array([np.count_nonzero(row) > 0
                                               for row in cc[:, colidx]])
                    e = fs['e'][nonzerorowmask][0]
                    c = cc[nonzerorowmask, :][0][colidx]
                    out += "{i:5d} & {e:>{efmt}} & \\\ \n".format(i=count, e=e,
                                                                  efmt=efmt)
        out += '\end{tabular}'
        return out

    def to_molpro(self, withpars=False, efmt="20.10f", cfmt="15.8f"):
        '''
        Return a string with the basis set in MOLPRO format.

        Args:
            withpars : bool
                A flag to indicate whether to wrap the basis with
                `basis={ }` string
            efmt : str
                Output format for the exponents, default: "20.10f"
            cfmt : str
                Output format for the contraction coefficients,
                default: "15.8f"

        Returns:
            res : str
                basis set string
        '''

        # reorder indices of cf if necessary without modiyfing the instance
        funs = deepcopy(self.functions)
        for shell, fs in funs.items():
            if not has_consecutive_indices(fs):
                funs[shell] = reorder_shell_to_consecutive(fs)

        res = ""
        for shell, fs in funs.items():
            exps = ", ".join(["{0:>{efmt}}".format(e, efmt=efmt).lstrip()
                              for e in fs['e']]) + '\n'
            res += "{s:>s}, {e:>s}, ".format(s=shell, e=self.element) + exps

            for icol, cf in enumerate(fs['cf']):
                if cf.size == 1:
                    coeffs = ", ".join(["{0:>{cfmt}}".format(cc,
                        cfmt=cfmt).lstrip() for cc in cf['cc']])
                    res += "c, {0:d}.{0:d}, ".format(cf['idx'][0] + 1) + \
                        coeffs + '\n'
                else:
                    # indices should be consecutive
                    coeffs = ", ".join(["{0:>{cfmt}}".format(cc,
                        cfmt=cfmt).lstrip() for cc in cf['cc']])
                    res += "c, {0:d}.{1:d}, ".format(cf['idx'].min() + 1,
                                                     cf['idx'].max() + 1) +\
                        coeffs + '\n'
        if withpars:
            res = 'basis={\n' + res + '}'
        return res

    def to_nwchem(self, efmt="20.10f", cfmt="15.8f"):
        '''
        Return a string with the basis set in NWChem format.

        Args:
          efmt : str
            string describing output format for the exponents,
            default: "20.10f"
          cfmt : str
            string describing output format for the contraction
            coefficients, default: "15.8f"

        Returns:
          res : str
            basis set string in NwChem format
        '''

        res = 'BASIS "ao basis" PRINT\n'
        for shell, fs in self.functions.items():
            # create an array of contraction coefficients for a given shell
            cc = self.contraction_matrix(shell)

            # select columns with more than 1 non zero coefficients
            nonzerocolmask = np.array([np.count_nonzero(col) > 1
                                       for col in cc.T])
            nonzerorowmask = np.array([np.count_nonzero(row) > 0
                                       for row in cc[:, nonzerocolmask]])
            if np.any(nonzerocolmask):
                res += "{e} {s}\n".format(e=self.element, s=shell)
                for expt, cfc in zip(fs['e'][nonzerorowmask],
                                     cc[np.ix_(nonzerorowmask, nonzerocolmask)]):
                    res += "{e:>{efmt}}{c}".format(e=expt, efmt=efmt, c="".join(["{0:{cfmt}}".format(c, cfmt=cfmt) for c in cfc])) + "\n"

            if np.any(np.logical_not(nonzerocolmask)):
                for colidx in np.where(np.logical_not(nonzerocolmask))[0]:
                    res += "{e} {s}\n".format(e=self.element, s=shell)
                    nonzerorowmask = np.array([np.count_nonzero(row) > 0
                                               for row in cc[:, colidx]])
                    e = fs['e'][nonzerorowmask][0]
                    c = cc[nonzerorowmask, :][0][colidx]
                    res += "{e:>{efmt}}{c:>{cfmt}}".format(e=e, efmt=efmt, c=c,
                                                           cfmt=cfmt) + "\n"
        return res + "END\n"

    def to_pickle(self, fname=None):
        '''Save the basis set in pickle format under the filename `fname`

        Args:
          fname : str
            File name
        '''

        if fname is None:
            fname = self.get_filename('pkl')

        with open(fname, 'wb') as fbas:
            pickle.dump(self, fbas)

    def print_functions(self, efmt="20.10f", cfmt="15.8f"):
        '''
        Return a string with the basis set.

        Args:
          efmt : str
            string describing output format for the exponents,
            default: "20.10f"
          cfmt : str
            string describing output format for the contraction
            coefficients, default: "15.8f"

        Returns:
          res : str
            basis set string
        '''

        res = ''
        for shell, fs in self.functions.items():
            # create an array with all the contraction coefficients
            # for a given shell
            res += "\n" + "{s} shell".format(s=shell).center(40, '=') + "\n"
            cc = self.contraction_matrix(shell)
            count = 0
            # select columns with more than 1 non zero coefficients
            nonzerocolmask = np.array([np.count_nonzero(col) > 1
                                       for col in cc.T])
            nonzerorowmask = np.array([np.count_nonzero(row) > 0
                                       for row in cc[:, nonzerocolmask]])
            if np.any(nonzerocolmask):
                res += 'Contracted:\n'
                for expt, cfc in zip(fs['e'][nonzerorowmask],
                                     cc[np.ix_(nonzerorowmask,
                                               nonzerocolmask)]):
                    count += 1
                    coeffrow = "".join(["{0:{cfmt}}".format(c, cfmt=cfmt)
                                        for c in cfc])
                    res += "{i:5d}{e:>{efmt}}{c}".format(i=count, e=expt,
                                                         efmt=efmt,
                                                         c=coeffrow) + "\n"

            if np.any(np.logical_not(nonzerocolmask)):
                res += 'Uncontracted:\n'
                for colidx in np.where(np.logical_not(nonzerocolmask))[0]:
                    count += 1
                    nonzerorowmask = np.array([np.count_nonzero(row) > 0
                                               for row in cc[:, colidx]])
                    e = fs['e'][nonzerorowmask][0]
                    c = cc[nonzerorowmask, :][0][colidx]
                    res += "{i:5d}{e:>{efmt}}{c:>{cfmt}}".format(i=count, e=e,
                                                                 efmt=efmt,
                                                                 c=c,
                                                                 cfmt=cfmt) + "\n"
        return res

    def __repr__(self):
        keys = ['name', 'element', 'family', 'kind']

        res = "<BasisSet(\n"
        for key in keys:
            res += "\t{k:<20s} = {v}\n".format(k=key, v=getattr(self, key))
        res += self.print_functions()
        res += ")>"
        return res

    def __str__(self):
        keys = ['name', 'element', 'family', 'kind']

        res = ''
        for key in keys:
            res += "{k:<20s} = {v}\n".format(k=key.capitalize(),
                                             v=getattr(self, key))
        res += 'Functions:\n'
        res += self.print_functions()
        return res


def has_consecutive_indices(shell):
    '''
    Check if all the contracted functions have consecutive indices

    Args:
        shell : dict
            Basis functions for a given shell asa dict with structure
            ``{'e' : np.array(), 'cf': [np.array(), np.array(), ...]}``
    '''

    for i, cf in enumerate(shell['cf']):
        if len(cf) > 1:
            if not np.all(np.sort(cf['idx'])[1:] -
                          np.sort(cf['idx'])[:-1] == 1):
                return False
    else:
        return True


def reorder_shell_to_consecutive(shell):
    '''
    Reorder the exponents so that the indices of the contracted functions
    have consecutive inidices.

    Args:
        shell : dict
            Basis functions for a given shell asa dict with structure
            ``{'e' : np.array(), 'cf': [np.array(), np.array(), ...]}``

    Returns:
        shell : dict
            Same shell as on input but with reordered exponents and relabelled
            contracted functions
    '''

    # get the index of the largest contracted function
    cfidx = np.argmax([len(x) for x in shell['cf']])

    # get the new indices where the ones corresponding to the contracted
    # function are consecutive
    oldidx = shell['cf'][cfidx]['idx']
    conidx = np.arange(shell['e'].size, dtype=int)
    restidx = np.setdiff1d(conidx, oldidx)
    newidx = np.concatenate((oldidx, restidx))

    funs = {'e': shell['e'][newidx], 'cf': []}

    newidxlist = newidx.tolist()

    for cf in shell['cf']:
        funs['cf'].append(np.array([(newidxlist.index(i), c) for i, c in cf],
                                   dtype=CFDTYPE))

    # check if all the contracted functions have consecutive indices
    if not has_consecutive_indices(funs):
        raise ValueError('cannot reorder functions, check #cf: {}'.format(i))

    return funs


def merge(first, other):
    '''
    Merge functions from two BasisSet objects

    Args:
      first : BasisSet
        First `BasisSet` object to merge
      other : BasisSet
        Second `BasisSet` object to merge

    Returns:
      our : BasisSet
        BasisSet instance with functions from `first` and `other` merged
    '''

    newf = deepcopy(first.functions)
    for oshell, ofs in other.functions.items():
        if oshell.lower() in first.functions.keys():
            exps, idxs, idxo = merge_exponents(first.functions[oshell]['e'], ofs['e'])
            newf[oshell]['e'] = exps
            for cf in newf[oshell]['cf']:
                cf['idx'] = idxs[cf['idx']]
            newcf = [np.copy(cf) for cf in ofs['cf']]
            for cf in newcf:
                cf['idx'] = idxo[cf['idx']]
            newf[oshell]['cf'].extend(newcf)
        else:
            newf[oshell] = deepcopy(ofs)

    return BasisSet(name=first.name, element=first.element, functions=newf)


def primitive_overlap(l, a, b):
    '''
    Calculate the overlap integrals for a given shell `l` and two sets of
    exponents

    .. math::

       S(\zeta_i, \zeta_j) = \\frac{2^{l + \\frac{3}{2}} \left(\zeta_{1} \zeta_{2}\\right)^{\\frac{l}{2} + \\frac{3}{4}}}{\left(\zeta_{1} + \zeta_{2}\\right)^{l + \\frac{3}{2}}}

    Args:
      l : int
        Angular momentum quantum number of the shell
      a (M,) : numpy.array
        First vector of exponents
      b (N,) : numpy.array
        Second vector of exponents

    Returns:
      out (M, N) : numpy.array
        Overlap integrals
    '''

    return np.power(2.0 * np.sqrt(np.outer(a, b)) / np.add.outer(a, b), l + 1.5)


def nspherical(l):
    '''
    Calculate the number of spherical components of a function with a given
    angular momentum value `l`.
    '''
    return 2 * l + 1


def ncartesian(l):
    '''
    Calculate the number of cartesian components of a function with a given
    angular momentum value *l*.
    '''
    return int((l + 1) * (l + 2) / 2)


def zetas2legendre(zetas, kmax):
    '''
    From a set of exponents (`zetas`), using least square fit calculate the
    expansion coefficients into the legendre polynomials of the order `kmax`.

    Args:
      kmax : int
        length of the legendre expansion
      zetas : list
        list of exponents (floats) to be fitted

    Returns:
      coeff : numpy.array
        numpy array of legendre expansion coeffcients of length kmax
    '''

    # exponents should be sorted in the ascending order
    zetas = sorted(zetas)
    c = np.asarray([1.0] * kmax, dtype=float)

    leg = np.polynomial.Legendre(c)
    a = np.zeros((len(zetas), kmax))

    for j in range(len(zetas)):
        for k in range(kmax):
            arg = (2.0 * (j + 1.0) - 2.0) / (len(zetas) - 1.0) - 1.0
            a[j, k] = leg.basis(k)(arg)

    return np.linalg.lstsq(a, np.log(zetas))[0]


def zetas2eventemp(zetas):
    '''
    From a set of exponents (`zetas`), using least square fit calculate the
    approximate $\\alpha$ and $\\beta$ parameters from the even tempered
    expansion.

    Args:
      zetas : list
        list of exponents (floats) to be fitted

    Returns:
      coeff : numpy.array
        numpy array of legendre expansion coeffcients of length kmax
    '''

    return min(zetas), np.power(max(zetas) / min(zetas), 1.0 / (len(zetas) - 1))


def generate_exponents(formula, nf, params):
    '''
    Generate a sequence of exponents from a specified formula

    Args:
      formula : str
        name of the sequence from which the exponents are generated, one of:
          - `et`, `even`, `eventemp`, `even tempered`
          - `wt`, `well`, `welltemp`, `well tempered`
          - `le`, `legendre`
          - `exp`, `exponents`
      nf : int
        number of exponents
      params : tuple
        parameters for the formula

    Returns:
      numpy.array with exponents
    '''

    if formula.lower() in ["et", "even", "eventemp", "even tempered"]:
        return eventemp(nf, params)
    elif formula.lower() in ["wt", "well", "welltemp", "well tempered"]:
        return welltemp(nf, params)
    elif formula.lower() in ["le", "legendre"]:
        return legendre(nf, params)
    elif formula.lower() in ["exp", "exponents"]:
        return np.sort(np.array(params))[::-1]
    else:
        raise ValueError('unknown sequence name: {}'.format(formula))


def get_num_params(ftuple):
    '''
    Get the number of parameter for a given 4-tuple desciribing functions

    >>> get_num_params(('s', 'et', 20, (0.4, 2.0)))
    2
    '''

    formula = ftuple[1]
    if formula.lower() in ["et", "even", "eventemp", "even tempered"]:
        return 2
    elif formula.lower() in ["wt", "well", "welltemp", "well tempered"]:
        return 4
    elif formula.lower() in ["le", "legendre"]:
        return ftuple[2]
    elif formula.lower() in ["exp", "exponents"]:
        return ftuple[2]
    else:
        raise ValueError('unknown sequence name: {}'.format(formula))


def eventemp(numexp, params):
    '''
    Generate a sequence of nf even tempered exponents accodring to
    the even tempered formula

    .. math::
       \\zeta_i = \\alpha \cdot \\beta^{i-1}

    Args:
        numexp : int
            Number fo exponents to generate
        params : tuple of floats
            Alpha and beta parameters
    Returns:
        res : numpy array
            Array of generated exponents (floats)
    '''

    if not isinstance(numexp, int):
        raise TypeError('"nf" variable should be of "int" type, got: {}'.format(type(numexp)))
    if len(params) != 2:
        raise ValueError('"params" tuple should have exactly 2 entries, got {}'.format(len(params)))

    alpha, beta = params
    zetas = alpha * np.power(beta, np.arange(numexp))
    return zetas[::-1]


def welltemp(numexp, params):
    '''
    Generate a sequence of nf well tempered exponents accodring to
    the well tempered fromula

    .. math::

       \\zeta_i = \\alpha \cdot \\beta^{i-1} \cdot \\left[1 + \\gamma \cdot \\left(\\frac{i}{N}\\right)^{\delta}\\right]

    Args:
        numexp : int
            Number fo exponents to generate
        params : tuple of floats
            Alpha, beta, gamma and delta parameters

    Returns:
        res : numpy.array
            Array of generated exponents (floats)
    '''

    if not isinstance(numexp, int):
        raise TypeError('"nf" variable should be of "int" type, got: {}'.format(type(numexp)))
    if len(params) != 4:
        raise ValueError('"params" tuple should have exactly 4 entries, got {}'.format(len(params)))

    alpha, beta, gamma, delta = params
    zetas = alpha * np.power(beta, np.arange(numexp)) * \
            (1 + gamma * np.power(np.arange(1, numexp + 1) / numexp, delta))
    zetas.sort()
    return zetas[::-1]


def legendre(numexp, coeffs):
    '''
    Generate a sequence of nf exponents from expansion in the orthonormal
    legendre polynomials as described in: Peterson, G. A. et.al J. Chem. Phys.,
    Vol. 118, No. 3 (2003), eq. (7).

    .. math::
       \ln \\zeta_i = \\sum^{k_{\max}}_{k=0} A_k P_k \\left(\\frac{2j-2}{N-1}-1\\right)

    Args:
        numexp : int
            Number fo exponents to generate
        params : tuple of floats
            Polynomial coefficients (expansion parameters)

    Returns:
        res : numpy array
            Array of generated exponents (floats)
    '''

    if not isinstance(numexp, int):
        raise TypeError('"nf" variable should be of "int" type, got: {}'.format(type(numexp)))
    if len(coeffs) < 1:
        raise ValueError('"coeffs" tuple should have at least 1 entry, got {}'.format(len(coeffs)))

    # special case for one function
    if len(coeffs) == 1:
        return [np.exp(coeffs[0])]

    poly = np.polynomial.legendre.Legendre(coeffs)
    zetas = [poly(((2.0 * (i + 1.0) - 2.0) / (numexp - 1.0)) - 1.0) for i in range(numexp)]
    return np.exp(zetas[::-1])


def splitlist(l, n):
    '''
    Split a list into sublists of size `n`
    '''

    if len(l) % n == 0:
        splits = len(l) // n
    elif len(l) % n != 0 and len(l) > n:
        splits = len(l) // n + 1
    else:
        splits = 1

    for i in range(splits):
        yield l[n * i:n * i + n]


def sliceinto(toslice, chunk_sizes):
    '''
    Slice a list ``toslice`` into chunks of sizes defined in ``chunk_sizes``

    >>> sliceinto(list(range(10)), (6, 4))
    [[0, 1, 2, 3, 4, 5], [6, 7, 8, 9]]
    '''

    if len(toslice) != sum(chunk_sizes):
        raise ValueError('cannot slice list, size mismatch {} != {}'.format(len(toslice),
                                                                            sum(chunk_sizes)))

    sidx = [sum(chunk_sizes[:i]) for i in range(0, len(chunk_sizes))]
    return [toslice[i:i + s] for i, s in zip(sidx, chunk_sizes)]


def xyzlist(l):
    '''
    Generate an array of :math:`l_x`, :math:`l_y`, :math:`l_z` components of
    cartesian gaussian with a given angular momentum value in canonical order.

    For exampe:
      - ``l = 0`` generates the result ``array([[0, 0, 0]])``
      - ``l = 1`` generates ``array([[1, 0, 0], [0, 1, 0], [0, 0, 1])``
      - etc.

    The functions are coded by triples of powers of *x*, *y*, *z*, namely
    ``[1, 2, 3]`` corresponds to :math:`xy^{2}z^{3}`.

    Args:
      l : int
        Angular momentum value

    Returns:
      out ((l+1)*(l+2)/2, 3) : numpy.array
        Array of monomial powers, where columns correspond to *x*, *y*, and *z*
        respectively and rows correspond to functions.
    '''

    ncart = (l + 1) * (l + 2) // 2
    out = np.zeros((ncart, 3), dtype=np.int32)
    index = 0
    for i in range(l + 1):
        for j in range(i + 1):
            out[index, :] = [l - i, i - j, j]
            index += 1
    return out


def zlmtoxyz(l):
    '''
    Generates the expansion coefficients of the real spherical harmonics in
    terms of products of cartesian components. Method based on [Schelegel1995]_

    .. [Schelegel1995] Schlegel, H. B., & Frisch, M. J. (1995). "Transformation between
       Cartesian and pure spherical harmonic Gaussians". International Journal
       of Quantum Chemistry, 54(2), 83–87.
       `doi:10.1002/qua.560540202 <http:www.dx.doi.org/10.1002/qua.560540202>`_

    Args:
      l : int
        Angular momentum value

    Returns:
      out :math:`((l+1)*(l+2)/2, 2*l + 1)` : numpy.array
        Expansion coefficients of real spherical harmonics in terms of
        cartesian gaussians
    '''

    ncart = (l + 1) * (l + 2) // 2
    nspher = 2 * l + 1

    out = np.zeros((ncart, nspher))

    cartc = xyzlist(l)

    m = 0
    for icart in range(ncart):
        kx = cartc[icart, 0]
        ky = cartc[icart, 1]
        kz = cartc[icart, 2]
        if kx + ky % 2 == 1:
            continue
        j = (kx + ky) / 2
        tmpc = factorial2(2 * kx - 1) * factorial2(2 * ky - 1) * factorial2(2 * kz - 1) / factorial2(2 * l - 1)
        tmpc = np.sqrt(tmpc) / factorial(j)
        for i in range(l // 2 + 1):
            if j > i:
                continue
            if kx % 2 == 1:
                continue
            k = kx // 2
            if k > j:
                continue
            tmpi = tmpc * factorial2(2 * l - 2 * i - 1) / factorial(l - 2 * i) / factorial(i - j) / 2.0**i
            if i % 2 == 1:
                tmpi = -tmpi
            out[icart, l] += binom(j, k) * tmpi

    for m in range(1, l + 1):
        for icart in range(ncart):
            kx = cartc[icart, 0]
            ky = cartc[icart, 1]
            kz = cartc[icart, 2]
            jj = kx + ky - m
            if jj % 2 == 1:
                continue
            if jj < 0:
                continue
            j = jj // 2
            tmpc = factorial2(2 * kx - 1) * factorial2(2 * ky - 1) * factorial2(2 * kz - 1) / factorial2(2 * l - 1)
            tmpc = np.sqrt(2.0 * tmpc * factorial(l - m) / factorial(l + m)) / factorial(j)
            for i in range((l - m) // 2 + 1):
                if j > i:
                    continue
                tmpi = tmpc * factorial2(2 * l - 2 * i - 1) / factorial(l - m - 2 * i) / factorial(i - j) / 2.0**i
                if i % 2 == 1:
                    tmpi = -tmpi
                for k in range(j + 1):
                    kk = kx - 2 * k
                    if kk < 0 or kk > m:
                        continue
                    tmpk = tmpi * binom(j, k) * binom(m, kk)
                    kkk = m - kk
                    if kkk % 4 == 0:
                        out[icart, m + l] += tmpk
                    elif kkk % 4 == 1:
                        out[icart, -m + l] += tmpk
                    elif kkk % 4 == 2:
                        out[icart, m + l] -= tmpk
                    else:
                        out[icart, -m + l] -= tmpk
    return out



# -*- coding: utf-8 -*-

#The MIT License (MIT)
#
#Copyright (c) 2014 Lukasz Mentel
#
#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:
#
#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.
#
#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.

'''
Module with convenience functions for parsing files
'''

from __future__ import print_function

import itertools
from collections import defaultdict


def contains(string, query):
    'Check if `string` contains `query`'
    return string.find(query) > -1


def locatelinenos(filename, tolocate):
    '''
    Given a file and a list of strings return a dict with string as keys
    and line numbers in which they appear a values.

    Args:
      filename : str
        Name of the file
      tolocate : list of tuples
        List of tuples with strings to find (queries) as first elements and
        integer offset values as second

    Returns:
      out : dict
        Dictionary whose keys are indices corresponding to item in input list
        and values are lists of line numbers in which those string appear

    TODO:
        - add option to ignore the case of the strings to search
    '''

    out = defaultdict(list)
    for lineno, line in enumerate(open(filename, 'r')):
        for idx, (query, offset) in enumerate(tolocate):
            if contains(line, query):
                out[idx].append(lineno + offset)
    return out


def getlines(filename, tolocate):
    '''
    Return the lines from the files based on `tolocate`

    Args:
      filename : str
        Name of the file
      tolocate : list of tuples
        List of tuples with strings to find (queries) as first elements and
        integer offset values as second

    Return:
    '''

    located = locatelinenos(filename, tolocate)

    if len(tolocate) == len(located):
        for k, v in located.items():
            if len(v) > 1:
                raise ValueError('multiple lines found for "{0}": {1}'.format(
                    tolocate[k][0], ', '.join([str(x) for x in v])))

        startlno = min(list(itertools.chain(*located.values())))
        endlno = max(list(itertools.chain(*located.values())))
        return getchunk(filename, startlno, endlno)
    else:
        # TODO: this needs to be corrected to be more informative
        raise ValueError('len(tolocate) != len(located): {0} != {1}'.format(
            len(tolocate), len(located)))


def getchunk(filename, startlno, endlno):
    '''
    Get a list of lines from a file between specified line numbers `startlno`
    and `endlno`.

    Args:
      filename : str
        Name of the file to process
      startlno : int
        Number of the first line to obtain
      endlno : int
        Number of the last line to obtain

    Returns:
      lines : list
        A list of lines from the file `filename` between line numbers\
        `startlno` and `endlno`
    '''

    fobj = open(filename, 'r')
    fileiter = iter(fobj)

    for _ in range(startlno):
        next(fileiter)

    return [next(fileiter) for _ in range(endlno - startlno)]


def take(seq, num):
    '''
    Iterate over a sequence `seq` `num` times and return the list of the
    elements iterated over.
    '''
    return [next(seq) for _ in range(num)]


def parsepairs(los, sep="="):
    '''
    Parse a given list of strings "los" into a dictionary based on
    separation by "sep" character and return the dictionary.
    '''

    out = []
    for line in los:
        if sep in line:
            (name, value) = line.split(sep)
            out.append((name.strip(), float(value)))
    return dict(out)


def sliceafter(seq, item, num):
    '''
    Return "num" elements of a sequence "seq" present after the item "item".
    '''

    it = iter(seq)
    for element in it:
        if item in element:
            return [next(it) for _ in range(num)]


def slicebetween(string, start, end):
    '''
    Return a slice of the `string` between phrases `start` and `end`.
    '''

    istart = string.index(start)
    iend = string[istart:].index(end)
    return string[istart + len(start):istart + iend]



# -*- coding: utf-8 -*-

import numpy as np
import os

from itertools import product

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from chemtools.pescan.model import Base, Atom, Dimer, Trimer, Tetramer


def grid_product(dicts):
    '''
    Given a dict whose item are lists return a generator over dict with
    all possible selections from the list (cartesian product)

    >>> options = {"number": [1,2,3], "color": ["orange","blue"] }
    >>> print list( my_product(options) )
    [ {"number": 1, "color": "orange"},
      {"number": 1, "color": "blue"},
      {"number": 2, "color": "orange"},
      {"number": 2, "color": "blue"},
      {"number": 3, "color": "orange"},
      {"number": 3, "color": "blue"}
    ]
    '''
    return (dict(zip(dicts, x)) for x in product(*dicts.itervalues()))


def expand_grids(grids=[], debug=False):

    '''Expand the grid definition and calculate all the grid points.'''

    res = list()
    for grid in grids:
        array = np.empty(0)
        for start, stop, num in grid:
            array = np.concatenate((array, np.linspace(start, stop, num)), axis=0)
        res.append(array)
    return res

# controller methods


def add_atom(session, name, basisset):

    atom = Atom(name=name,
                abspath=None,
                output_name=None,
                basisset=basisset,
                hf_energy=None,
                ci_energy=None,
                cc_energy=None,
                status=None)
    session.add(atom)
    session.commit()
    return True


def add_trimer(session, name, basisset, ram, raa, gamma):
    '''
    Add a Trimer to the database.
    '''

    trimer = Trimer(
        name=name,
        abspath=None,
        output_name=None,
        basisset=basisset,
        hf_energy=None,
        ci_energy=None,
        cc_energy=None,
        status=None,
        r_atom_mol=ram,
        gamma=gamma,
        r_atom1_atom2=raa)
    session.add(trimer)
    session.commit()
    return True


def add_tetramer(session, name, basisset, r_mol1_mol2, r_mol1, r_mol2, phi_1,
                 phi_2, gamma):
    '''
    Add a Tetramer to the database.
    '''

    tetramer = Tetramer(
        name=name,
        basisset=basisset,
        r_mol1_mol2=r_mol1_mol2,
        r_mol1=r_mol1,
        r_mol2=r_mol2,
        gamma=gamma,
        phi_1=phi_1,
        phi_2=phi_2)
    session.add(tetramer)
    session.commit()
    return True


def add_dimer(session, name, basisset, raa):
    '''
    Loop over all jobs and return a list of dicts with job info

    Args:
      session:
        database session object (connection)
      name : str
        name of the system
      basissets : list
        list of strings with basis set names
      grids : numpy.array
        1-D numpy array with the values of the internuclear distances

    '''

    dimer = Dimer(name=name,
                  abspath=None,
                  output_name=None,
                  basisset=basisset,
                  hf_energy=None,
                  ci_energy=None,
                  cc_energy=None,
                  status=None,
                  r_atom1_atom2=raa)
    session.add(dimer)
    session.commit()
    return True


def create_dirs(session, modelobj=None, workdir=os.getcwd()):
    '''
    Create the directory tree associated with a given table.
    '''

    # if workdir doesn't exist create it
    # raise exception when modelobj is None

    os.chdir(workdir)

    molecules = session.query(modelobj.name).distinct().all()
    for mol in molecules:
        os.mkdir(mol.name)
        os.chdir(mol.name)
        basis = session.query(modelobj.basisset).filter(modelobj.name == mol.name).distinct().all()
        for bas in basis:
            os.mkdir(bas.basisset)
            jobpath = os.path.join(os.getcwd(), bas.basisset)
            for entry in session.query(modelobj).filter(modelobj.name == mol.name).filter(modelobj.basisset == bas.basisset).all():
                entry.abspath = jobpath
                session.add(entry)
        os.chdir("..")
    session.commit()

# database connection methods

def new_db(path):
    '''
    Create a new database under path and return the session.
    '''
    engine = create_engine("sqlite:///{path:s}".format(path=path), echo=False)
    Base.metadata.create_all(engine)
    db_session = sessionmaker(bind=engine)
    return db_session()

def get_session(dbpath):
    '''
    Connect to a database under the name stored in "path" and return the
    session.
    '''

    if not os.path.exists(dbpath):
        raise OSError("database {0} does not exist".format(path))

    engine = create_engine("sqlite:///{path:s}".format(path=dbpath), echo=False)
    db_session = sessionmaker(bind=engine)
    return db_session()



# -*- coding: utf-8 -*-

from __future__ import print_function

from argparse import ArgumentParser
import os
import sys

from chemtools.basisset import BasisSet


def bsprint():
    '''
    CLI script to read a basis set from a file in internal format and print
    in a specified format
    '''

    parser = ArgumentParser()
    parser.add_argument("file",
                        help="file name with a pickled BasisSet object")
    parser.add_argument("-f", "--format",
                        choices=["cfour", "dalton", "gamessus", "gaussian",
                                 "json", "molpro", "nwchem"],
                        default="default")
    args = parser.parse_args()

    bs = BasisSet.from_pickle(args.file)

    if args.format == 'default':
        print(str(bs))
    else:
        writer = "to_" + args.format
        method = getattr(bs, writer)
        print(method())


def bsconvert():
    '''
    CLI script to convert between different basis set formats
    '''

    parser = ArgumentParser(description='Convert basis set between formats of different programs')
    parser.add_argument("filename",
                        help="file name with a basis set, default='pickle'")
    parser.add_argument("-from",
                        "--inputformat",
                        choices=["gamessus", "gaussian", "json", "molpro",
                                 "pickle"],
                        help="Basis set input format",
                        default="pickle")
    parser.add_argument("-to",
                        "--outputformat",
                        choices=["cfour", "dalton", "gamessus", "gaussian",
                                 "json", "molpro", "nwchem", "pickle"],
                        help="Basis set output format",
                        default="molpro")
    parser.add_argument('-o', '--output', help='name of the output file')
    args = parser.parse_args()

    name = os.path.splitext(args.filename)[0]

    if args.inputformat == "pickle":
        bsets = BasisSet.from_pickle(args.filename)
    elif args.inputformat == "json":
        bsets = BasisSet.from_json(args.filename)
    else:
        bsets = BasisSet.from_file(fname=args.filename, fmt=args.inputformat,
                                   name=name)

    if args.outputformat == "pickle":
        if isinstance(bsets, dict):
            for elem, bset in bsets.items():
                bset.to_pickle(name + '_' + elem + '.pkl')
        elif isinstance(bsets, BasisSet):
            bsets.to_pickle(name + '_' + bsets.element + '.pkl')
    else:
        writer_name = "to_" + args.outputformat
        writer_objs = []
        if isinstance(bsets, dict):
            for elem, bset in bsets.items():
                writer_objs.append(getattr(bset, writer_name))
        elif isinstance(bsets, BasisSet):
            writer_objs.append(getattr(bsets, writer_name))
        else:
            raise ValueError('Something went wrong')

        if args.output:
            fobj = open(args.output, 'w')
        else:
            fobj = sys.stdout

        for writer in writer_objs:
            print(writer(), file=fobj)



# -*- coding: utf-8 -*-

#The MIT License (MIT)
#
#Copyright (c) 2014 Lukasz Mentel
#
#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:
#
#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.
#
#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.

'Orbitals class'

import numpy as np
import pandas as pd

from chemtools.calculators.gamessus import GamessLogParser, GamessDatParser
from chemtools.calculators.gamessreader import DictionaryFile, tri2full

class Orbitals(pd.DataFrame):
    '''
    A convenience class for handling GAMESS(US) orbitals.
    '''

    def __init__(self, *args, **kwargs):

        super(Orbitals, self).__init__(*args, **kwargs)

    @classmethod
    def from_files(cls, name=None, logfile=None, dictfile=None, datfile=None):
        '''
        Initialize the `Orbitals` instance based on orbital information parsed from the `logfile`
        and read from the `dictfile`.

        Args:
            name : str
                One of `hf` or `ci`
            logfile : str
                Name of the GAMESS(US) log file
            dictfile : str
                Name of the GAMESS(US) dictionary file .F10
            datfile : str :
                Name of the GAMESS(US) dat file
        '''

        if name == 'hf':
            evec_record = 15
            eval_record = 17
            syml_record = 255
        elif name == 'ci':
            evec_record = 19
            eval_record = 21
            syml_record = 256
        elif name == 'local':
            pass
        else:
            raise ValueError('name should be one either "hf", "ci" or "local"')

        if name in ['hf', 'ci']:
            # parse the number of aos and mos
            glp = GamessLogParser(logfile)
            nao = glp.get_number_of_aos()
            nmo = glp.get_number_of_mos()

            # read the relevant record from the dictfile
            df = DictionaryFile(dictfile)
            # read the orbitals
            vector = df.read_record(evec_record)
            vecs = vector[:nao*nmo].reshape((nao, nmo), order='F')
            # read the eigenvectors
            evals = df.read_record(eval_record)
            evals = evals[:nmo]
            # read symmetry labels
            symlab = df.read_record(syml_record)
            symlab = symlab[:nmo]

            data = dict([
                ('symlabels', [s.strip() for s in symlab]),
                ('eigenvals', evals),
                ('gindex', range(1, nmo + 1)),
            ])
        elif name == 'local':
            gdp = GamessDatParser(datfile)
            vecs = gdp.get_orbitals(name)
            nao, nmo = vecs.shape

            data = dict([('gindex', range(1, nmo + 1)),])

        dataframe = cls(data=data)

        dataframe.nao = nao
        dataframe.nmo = nmo
        dataframe.name = name
        dataframe.logfile = logfile
        dataframe.dictfile = dictfile
        dataframe.datfile = datfile
        dataframe.coeffs = pd.DataFrame(data=vecs)

        return dataframe

    def assign_lz_values(self, decimals=6, tolv=1.0e-2):
        '''
        Determine the eigenvalues of the Lz operator for each nondegenerate or
        a combination of degenerate orbitals and assign them to `lzvals` column
        in the DataFrame

        The Lz integrals over atomic orbitals are read from the dictionary file
        record No. 379.

        Args:
          decimals : int
            Number of decimals to keep when comparing float eigenvalues
          tolv : float
            Threshold for keeping wights of the eiegenvalues (squared eigenvector components)

        WARNING:
            currently this only works if the eigenvalues on input are sorted
        '''

        lzmap = {0: 'sigma', 1: 'pi', 2: 'delta', 3: 'phi'}

        df = DictionaryFile(self.dictfile)
        # read Lz integrals from the dictionary file
        lzvec = df.read_record(379)
        # expand triangular to square matrix
        lzints = tri2full(lzvec, sym=False)

        unq, indices = check_duplicates(self.eigenvals, decimals=decimals)

        out = []

        mos = self.coeffs.values

        for IG, (i, j) in enumerate(indices, start=1):
            # transform the integrals to the MO basis
            lzmo = np.dot(mos[:, i:j].T, np.dot(lzints, mos[:, i:j]))

            # tranform to complex type
            lzcplx = np.zeros_like(lzmo, dtype=np.complex128)
            lzcplx.imag = -lzmo

            # diagonalize the lzcplx matrix
            w, v = np.linalg.eig(lzcplx)

            wij2 = np.multiply(np.abs(v), np.abs(v))
            #x, y = np.where(wij2 > tolv)
            idxs = [np.nonzero(row > tolv)[0] for row in wij2]
            for row, comp in enumerate(idxs):
                out.append([(IG, w.real[m], wij2[row, m]*100.0) for m in comp])

        evs = [np.unique(np.round(np.abs(np.array([x[1] for x in row])), decimals=1)) for row in out]
        lzvals = [int(x[0]) if x.size == 1 else np.NaN for x in evs]

        self['lzvals'] = [int(x[0]) if x.size == 1 else np.NaN for x in evs]
        self['lzlabels'] = self['lzvals'].map(lzmap)
        return out

    def lzmos(self, ETOLLZ=1.0e-6, TOLV=1.0e-2):
        '''
        A python rewrite of the GAMESS(US) LZMOS subroutine (from symorb.src) for analyzing the Lz composition of the
        orbtials

        Args:
            ETOLLZ : float
            TOLV : float

        '''

        lzmap = {0: 'sigma', 1: 'pi', 2: 'delta', 3: 'phi'}

        df = DictionaryFile(self.dictfile)
        # read Lz integrals from the dictionary file
        lzvec = df.read_record(379)
        # expand triangular to square matrix
        lzints = tri2full(lzvec, sym=False)

        mos = self.coeffs.values
        ev = self.eigenvals.values
        out = []

        L = 0
        ILAST = 0
        IG = 0
        N = 1
        for I in range(1, self.nmo + 1):
            EI = 0
            if I < self.nmo: EI = ev[I]

            if (abs(ev[I - 1] - EI) > ETOLLZ or I > self.nmo):

                IG += 1
                # transform the integrals to the MO basis
                lzmo = np.dot(mos[:, ILAST:ILAST+N].T, np.dot(lzints, mos[:, ILAST:ILAST+N]))
                # tranform to complex type
                lzcplx = np.zeros_like(lzmo, dtype=np.complex128)
                lzcplx.imag = -lzmo

                # diagonalize the lzcplx matrix
                w, v = np.linalg.eig(lzcplx)

                for j in range(v.shape[0]):
                    m = -1
                    temp = []
                    for k in range(v.shape[1]):
                        wij = np.abs(v[j, k])
                        if wij*wij > TOLV:
                            m += 1
                            temp.append(wij*wij*100.0)
                    out.append([(IG, w.real[mi], temp[mi]) for mi in range(m+1)])

                ILAST = I
                N = 1
            else:
                N = N + 1

        # should check here if the length of the evs is the same as the number of rows of the dataframe
        # if somethings wrong probably the ETOLLZ should be adjusted
        evs = [np.unique(np.round(np.abs(np.array([x[1] for x in row])), decimals=1)) for row in out]
        lzvals = [int(x[0]) if x.size == 1 else np.NaN for x in evs]

        self['lzvals'] = [int(x[0]) if x.size == 1 else np.NaN for x in evs]
        self['lzlabels'] = self['lzvals'].map(lzmap)
        self['ig'] = [x[0][0] for x in out]
        return None

    def fragment_populations(self):
        '''
        Calculate the fragment populations and assign each orbital to a fragment
        '''

        glp = GamessLogParser(self.logfile)
        self.aoinfo = pd.DataFrame(data=glp.get_ao_labels(orbs=self.name + ' orbs'))

        idx1 = self.aoinfo.loc[self.aoinfo['center'] == 1, 'index'].values
        self['center 1 pop'] = (self.coeffs.loc[idx1, :]**2).sum()
        idx2 = self.aoinfo.loc[self.aoinfo['center'] == 2, 'index'].values
        self['center 2 pop'] = (self.coeffs.loc[idx2, :]**2).sum()

        self['fragment'] = np.where(self['center 1 pop'] > self['center 2 pop'], 1, 2)

def check_duplicates(a, decimals=6):
    '''
    This funciton assumes that the array `a` is sorted

    http://stackoverflow.com/questions/25264798/checking-for-and-indexing-non-unique-duplicate-values-in-a-numpy-array
    http://stackoverflow.com/questions/5426908/find-unique-elements-of-floating-point-array-in-numpy-with-comparison-using-a-d
    '''

    unq, unq_idx, unq_cnt = np.unique(np.round(a, decimals=decimals), return_inverse=True, return_counts=True)

    indices = np.zeros((len(unq), 2), dtype=np.int)
    for ig in range(len(unq)):
        idx = np.nonzero(unq_idx == ig)[0]
        indices[ig, :] = [np.min(idx), np.max(idx) + 1]

    return unq, indices
